/******************************************************************************
*
*       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*       SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*       FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*       FOR A PARTICULAR PURPOSE.
*
*       (c) Copyright 2005-2006 Xilinx Inc.
*       All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xtemac_end_adapter.c
*
* The Xilinx Ethernet Tri-Mode 10/100/1000 Adapter component. This component
* interfaces between the XTemac component and VxWorks 5.x or 6.x.  The adapter
* plugs into the VxWorks MUX interface. It follows the Enhanced Network Driver
* (END) model.
*
* This driver supports multiple frame transfer modes depending on what features
* have been built into the HW:
*
*   - FIFO Direct interrupt driven mode: Processor is used to access the FIFOs
*     directly. This mode is not suitable for 1 GB links due to lower
*     throughput capability.
*
*   - FIFO Direct polled mode operation is available for system mode debugging
*     with the WDB agent.
*
*   - Scatter Gather DMA (SGDMA) interrupt driven mode: A buffer descriptor
*     based DMA engine transfers multiple frames automatically to/from the
*     FIFOs.
*
* SGDMA mode can be further enhanced with the Data Realignment Engine (DRE) on
* the transmit channel and checksum offloading. DRE prevents the driver from
* copying frames passed down from the stack into a properly aligned buffer.
* Checksum offloading can boost throughput by allowing HW to calculate the TCP
* or UDP checksum.
*
* Note that when HW is configured for SGDMA, there is no polled mode available.
*
*
* <b>INIT STRING</b>
*
* The init string is passed into the driver at intitialization time. The string
* consists of several fields that instruct the driver on how to load itself
* and configure the device. Numeric fields may be specified in octal, decimal,
* or hexidecimal notation.
*
* The init string consists of the following fields: <pre>
* 
* <VxUnitNumber>:<DeviceId>:<MacAddrPtr>:<BufBaseAddr>:
* <BufBytes>:<BdBaseAddr>:<BdBytes>:<RxBdCnt>:<TxBdCnt><CacheLineSize>:
* <IntrId>:<MTU>
*
* where:
*
*   <VxUnitNumber> is a VxWorks-specific identifier for this driver. The unit
*   number will be prepended by muxEndLoad (a VxWorks muxLib function).
*
*   <DeviceId> is the unique identifier for the Ethernet controller being
*   initialized. This ID is assigned by the platform generator tool.
*
*   <MacAddrPtr> is the address of the first byte of the 6-byte MAC address.
*
*   <BufBaseAddr> & <BufBytes> defines an area of memory to be used to store
*   receive buffers and (if needed) transmit buffers copied from the stack.
*   The base address will be adjusted if it is not aligned correctly. If 
*   BufBaseAddr is "0", the adapter will allocate BufBytes of cache safe
*   memory using the cacheDmaMalloc routine. If BufBaseAddr is non-0, then it
*   is assumed the memory region is non-cached (if it turns out the memory
*   (region is cached then the driver will not work). If BufBytes is "0" then
*   a default number of buffers. XTEA_DFT_BUFS will be allocated. If BufBytes
*   is non "0", then it must be large enough to contain at least XTEA_MIN_BUFS.
*   A load error is generated if BufBytes is "0" but BufBaseAddr is not.
*
*   <BdBaseAddr> & <BdBytes> defines an area of cache safe memory to be used 
*   to store SGDMA buffer descriptors. The number of descriptors is defined by
*   the <RxBdCnt> and <TxBdCnt> fields. BdBaseAddr  must be aligned to
*   DESC_ALIGNMENT. If BdBaseAddr is "0", the driver will allocate
*   BufBytes of memory using the cacheDmaMalloc routines. If BdBytes is also
*   "0" then enough memory will be allocated to contain all BDs. If BdBytes
*   is non "0", then it must be large enough to contain all BDs. An error is
*   generated when 1) if BdBytes is "0" but BdBaseAddr is not; 2) BdBytes is
*   not large enough to contain all descriptors; 3) The memory area is not cache
*   coherent. These fields are ignored if the HW is not SGDMA capable.
*
*   <RxBdCnt> is the number of receive descriptors to use for SGDMA. If
*   specified as 0, then the system will default to XTEA_DFT_RXBDS (see
*   xtemac_end_adapter.h). This field is ignored if the HW is not SGDMA capable.
*
*   <TxBdCnt> is the number of transmit descriptors to use for SGDMA. If
*   specified as 0, then the system will default to XTEA_DFT_TXBDS (see
*   xtemac_end_adapter.h). This field is ignored if the HW is not SGDMA capable.
*
*   <CacheLineSize> is the size, in bytes, of the processor data cache-line.
*   It is used by the driver when allocating receive buffers so that each
*   buffer is cache-line aligned. If this value is less than 8 bytes, then a
*   default of 8 is used.
*
*   <IntrId> is the interrupt ID for the device. The interrupt ID is used when
*   the adapter connects the driver's interrupt handler to the interrupt
*   controller.  The ID is typically a value between 0 and 31, inclusive - but
*   it's value depends on what the interrupt controller expects when connecting.
*
*   <MTU> tells the driver how to set the Maximum Transmission Unit. If this
*   value is greater than 1500, then jumbo frame capability is enabled in the
*   hardware. This value should be no larger than XTEA_MAX_MTU. If specified
*   as 0, then MTU will default to 1500 bytes. When using jumbo frames, be sure
*   the device's packet FIFOs are large enough to fully contain at least one
*   frame. Packet FIFOs that are too small may cause the HW to hang.
* </pre>
*
*
* <b>CACHE COHERENCY</b>
*
* In all modes, the user may specify buffer space in cached memory. The driver
* will flush/invalidate caches when appropriate.
*
* For SGDMA modes, descriptor space requires read and write coherent memory. This
* means that the data cache may not be enabled for the memory region defined by
* <BdBaseAddr> and <BdBytes> fields of the InitString.
*
* During initialization, if the driver detects that memory coherency cannot be
* guaranteed for buffer descriptors, then the driver will refuse to load. This
* decision is based on the state of the cache at the time the driver is loaded.
* If the cache is disabled when this driver is loaded, then enabled at some
* later time, then memory coherency is lost with unpredictable results.
*
*
* <b>ALIGNMENT REQUIREMENTS AND BUFFER COPYING</b>
*
* The driver allocates buffer memory for the receive channel to meet the
* alignment requirements of the HW and to provide the best performance.
*
* Transmit channel buffers may or may not require special handling. Each
* mode is described below.
*
* FIFO Direct:
*
* Transmit frames passed down from the stack not on a BUF_ALIGNMENT
* alignment boundary will take longer to transfer.
*
* SGDMA without Tx DRE:
*
* HW alignment cannot be guaranteed by packets handed down by the stack. As a
* result, this driver will copy all packets to an aligned buffer prior to
* submitting them to the DMA engine. This stack-to-driver copy operation will
* impose a throughput performance penalty.
*
* During init time, the driver will divide up the pool of buffers into Rx and
* Tx groups with the Tx group allocated solely for stack-to-driver copy
* operations. 
* 
* Buffer descriptor memory area must be specified by the InitString on at least
* a DESC_ALIGNMENT boundary.
*
* SGDMA with Tx DRE:
*
* Transmit frames passed down from the stack can be on any alignment without
* performance penalties.
*
* A stack-to-driver copy will occur if the mblk chain is longer than
* XTEA_SG_MAX_FRAGS fragments.
*
* Buffer descriptor memory area must be specified by the InitString on at least
* a DESC_ALIGNMENT boundary.
*
*
* <b>BUFFER ORGANIZATION</b>
*
* Buffers allocated by this driver are sized to take into account the network
* MTU, alignment, and overhead required by VxWorks netBufLib. These factors
* will combine to create a buffer whose size is slighly larger than largest
* frame possible for the given MTU.
*
* For example, using an MTU of 1500 bytes, the maximum sized ethernet frame
* becomes 1518 bytes. But when alignment and overhead are taken into account,
* this may become 1524 or more bytes depending on what alignment is specified
* by the associated InitString field.
*
* The diagram below shows how buffers are laid out in memory. Note that the
* front and back end of each buffer is aligned. This is done so that cache
* flush/invalidate operations do not affect adjacent buffers or netBufLib
* overhead bytes for DMA equipped devices.
*
* <pre>
*                      +------------------+
*                      |netBufLib overhead|
*                 +--> +------------------+ <-- CacheLine aligned
*    netBufLib    |    |                  |
*     cluster ----|    |     Buf #1       |
*      size       |    |                  |
*                 |    +------------------+ <--+ 
*                 |    |     Filler       |    |  sized to fit within a cacheline
*                 +--> +------------------+    |
*                      |netBufLib overhead|    |
*                      +------------------+ <--+--- CacheLine aligned
*                      |                  |
*                      |     Buf #2       |
*                      |                  |
*                      +------------------+
*                      |     Filler       |
*                      +------------------+
*                              .
*                              .
* </pre>
*
* The driver will always allocate its buffer space based on what is provided
* by the InitString. The buffers are distributed between the receive and
* transmit channels. If the HW options are such that stack-to-driver copying
* is not needed, then all buffers are given to the receive channel.
*
* In SGDMA mode without Tx DRE, stack-to-driver copying is required due to HW
* alignment restrictions. The distribution of buffers is controlled by the
* XTEA_COPY_BUFS_RATIO. For example, if the total number of buffers allocated
* is 100 and the constant is set to 1, then 50 will be used for receive and
* 50 for transmit. A setting of 3, would lead to 75 receive buffers and 25
* transmit buffers. 
*
* In SGDMA mode with Tx DRE, stack-to-driver copying is required to deal with
* highly fragmented mblk chains or when the number of fragments is greater
* than the number of available descriptors. In this case XTEA_MAX_DEFRAG_BUFS
* is used to reserve a fixed number of buffers to coalesce fragments into prior
* to sending them. The upper limit on the number of fragments the driver will
* manage without copying them by default is XTEA_SG_MAX_FRAGS.
* 
* <b>IOCTL</b>
*
* In addition to the standard set of ioctl commands required of END drivers,
* this driver implements custom commands to control several advanced hardware 
* features that can be used to tune the device to the needs of the user system.
*
* <pre>
*    #include <muxLib.h>
*    #include <endLib.h>
*    #include <xtemac.h>
*
*    #define IOCTL(Eptr, Cmd, Data) \
*        ((*Eptr->pFuncTable->ioctl)(Eptr, Cmd, (caddr_t)Data))
*
*    void DoIoctl(int Command, caddr_t Data)
*    {
*       END_OBJ* Eptr;
*       int Result;
*
*       Eptr = endFindByName(XTE_DEVICE_NAME, UnitNumber);
*       Result = IOCTL(Eptr, Command, Data);
*    }
* </pre>
*
*
* <b>INTERRUPT COALESCING</b>
*
* When configured with scatter-gather DMA, the frequency of interrupts can be
* controlled with interrupt coalescing features of the DMA engine. Reducing the
* overhead of interrupt processing can lead to significant increases to
* performance. For a more complete discussion of these feature, see xtemac.h.
*
* Tweaking interrupt coalescing settings to get the best performance depends on
* the type of network traffic patterns the end user system will process. For
* example, performance may be enhanced by setting a high transmit packet
* threshold value when sending UDP packets in a bulk data transfer scenario.
* This same setting, however, may reduce performance with TCP packets utilizing
* small window sizes.
* 
* By default both the Rx and Tx packet thresholds default to 1 with the waitbound
* timer disabled (set to 0). This provides one interrupt per packet received or
* transmitted.
*
* To change the settings, use ioctl commands XTEA_EIOCSPKTTW and XTEA_EIOCGPKTTW
* as follows:
* <pre>
*    XTemacEnd_SgPktIoctlData PktData;
*    END_OBJ* Eptr = endFindByName(XTE_DEVICE_NAME, UnitNumber);
*
*    // Send the data to the device
*    PktData.RxPktWaitbound = 1;
*    PktData.TxPktWaitbound = 5;
*    PktData.RxPktThresholdCnt = 8;
*    PktData.TxPktThresholdCnt = 8;
*    IOCTL(Eptr, XTEA_EIOCSPKTTW, &PktData);
*
*    // Retrieve the data from the device
*    IOCTL(Eptr, XTEA_EIOCGPKTTW, &PktData);
* </pre>
*
*
* <b>CHECKSUM OFFLOADING FOR UDP & TCP PACKETS</b>
*
* HW can be built with checksum offloading on the transmit channel, the receive
* channel, or both channels. The HW implementation covers the UDP/TCP header and
* payload portion of the packet. IP and IP pseudo header checksum related
* calculations are left to the stack.
*
* <b>LINK CONFIGURATION</b>
*
* No PHY setup is performed by this driver. Users can communicate with the PHY
* if a communications path (MII, GMII, SGMII, RGMII, or 1000 Base-T) is present
* in HW.
*
* To use the Driver API to set/get MII registers the following algorithm
* can be used:
*
* <pre>
*   #include "xtemac.h"
*
*   int Key;
*   XTemac *TemacPtr = XTemacEndAdapter_GetDriver(UnitNumber);
*
*   if (TemacPtr == NULL)
*   {
*       // End driver not loaded
*   }
*
*   Key = intLock();
*   Status = XTemac_PhyRead(TemacPtr, PhyAddr, PhyReg);
*   Status = XTemac_PhyWrite(TemacPtr, PhyAddr, PhyReg, PhyData);
*   intUnlock(Key);
* </pre>
*
* The MAC-PHY clock is independent of the PHY's current link state. The TEMAC HW
* does not provide feedback to automatically set this clock. It is up to the user
* to query the PHY to determine it's link speed and relay that data to the
* TEMAC. Relaying this is done with an ioctl command:
*
* <pre>
*    END_OBJ* Eptr = endFindByName(XTE_DEVICE_NAME, UnitNumber);
*
*    // Set to 1000 Mbps
*    IOCTL(Eptr, XTEA_EIOCSMPSPD, 1000);
*
*    // Set to 100 Mbps
*    IOCTL(Eptr, XTEA_EIOCSMPSPD, 100);
*
*    // Set to 10 Mbps
*    IOCTL(Eptr, XTEA_EIOCSMPSPD, 10);
* </pre>
*
* <b>DEBUGGING</b>
*
* This adapter is instrumented with debugging code throughout. This code is
* compiled or ignored based on how XMAC_DEBUG is defined in this file. 
* When enabled, it can be controlled by global variable xtemac_debug also 
* defined in this file. Setting this variable can enable/disable debugging 
* messages in certain contexts. For example, it can be set to cause debug
* messages only from transmit routines. Since xtemac_debug is a globallly 
* defined integer, it can be set to an initial value at compile time or 
* changed on the fly at runtime through a debugger or the Tornado host/target
* shell. To use this debugging capability, you must INCLUDE_LOGGING in your bsp.
*
*
* <b>DRIVER ERRATA AND LIMITATIONS</b>
*
* 1) Message "temac: warning - MII clock is defaulted to <X> Mbps" appears
*    when device is loaded.
*
*    This warning is printed to tell the user a default PHY data clock has
*    been selected. The MAC core's data clock must be in sync with the link
*    speed the PHY has negotiated to (see LINK CONFIGURATION section). By
*    default this is set by the driver 100 Mbps if using MII or 1000 for
*    any other MII type. If, however, the actual negotiated is different then
*    the MAC will fail to pass traffic. An ioctl command is available to set
*    the correct data clock.
*
* 2) No traffic being passed even though links appear to be up.
*
*    One possible cause is related to errata 1. In this case, the MAC's data
*    clock is not in sync with the PHY. The user must access the PHY to
*    determine the negotiated speed (see section LINK CONFIGURATION). 
*
* 3) Message "temac: Descriptor space requires cache coherent memory" appears
*    when device is loaded. No traffic is passed.
*
*    This occurs when the init string specifies a descriptor space that the 
*    driver has detected as being data cached. The solution is to relocate
*    descriptor memory space to a non-cached area or disable the data cache.
*
* 4) The receive side DRE HW feature is not utilized because this driver 
*    automatically aligns receive frames along cache lines.
*
* 5) FCS stripping must be enabled when using the checksum offload feature
*    on the receive channel. If disabled, then HW will include the 4 byte CRC
*    at the end of the frame in its checksum computation. This will lead to
*    all UDP and TCP frames larger than 64 bytes being dropped due to incorrect
*    checksums.
*
* 6) FCS auto-insertion must be enabled when using the checksum offload feature
*    on the transmit channel. If disabled, the 4 byte CRC will be included in
*    the checksum calculations resulting in an incorrect checksum.
*
* 7) Checksum offloading is not supported when VLAN is enabled.
*
* 8) In a dual TEMAC system, MII access of one TEMAC blocks communication to
*    the other until the MII cycle completes. The blocking mechanism must be
*    enforced by SW (typically by locking out interrupts). Failure to block
*    access to the other TEMAC will result in system instability.
*
* @note
*
* Xilinx drivers are typically composed of two components, one is the driver
* and the other is the adapter.  The driver is independent of OS and processor
* and is intended to be highly portable.  The adapter is OS-specific and
* facilitates communication between the driver and the OS.
*
* This component does not adhere to the VxWorks Network Protocol Toolkit (NPT)
* driver model.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- --------------------------------------------------------
* 1.00a rmm  06/24/05 First release
* 2.00a rmm  11/21/05 Removed simple DMA mode, added checksum offload for
*                     VxWorks 6.x, HW fixes for MII and multicast, removed
*                     XTEA_EIOCSSGEND and XTEA_EIOCGSGEND ioctl commands
*       rmm  05/31/06 CR 228521: Fixed to use correct cache flushing macro in
*                     SgSend.
* </pre>
*
******************************************************************************/

/***************************** Include Files *********************************/
#include "xtemac_end_adapter.h"
#include "xtemac_l.h"
#include "xdmav3_l.h"
#include "xparameters.h"
#include "xstatus.h"
#include "config.h"

#include <string.h>
#include <cacheLib.h>
#include <stdlib.h>
#include <stdio.h>
#include <net/if.h>
#include <netLib.h>
#include <m2Lib.h>
#include <intLib.h>
#include <logLib.h>
#include <taskLib.h>
#include <sysLib.h>
#include <etherLib.h>
#include <etherMultiLib.h>


/************************** Constant Definitions *****************************/

XTemac *macptr;

/*
 * #define XMAC_DEBUG to turn debugging on. Once enabled, use the next block
 * of constant to control which calls to the DEBUG_LOG macro produce output. 
 * Set the xtemac_debug variable with one or more of these constants OR'd
 * together to control what is output.
 */
#undef XMAC_DEBUG                          

#ifdef XMAC_DEBUG
#  warning XMAC_DEBUG is enabled
#  define DEBUG_ERROR     0x0001    /* error condition messages */
#  define DEBUG_TX        0x0002    /* tx routines */
#  define DEBUG_RX        0x0004    /* rx routines */
#  define DEBUG_IOCTL     0x0008    /* ioctl calls */
#  define DEBUG_MISC      0x0010    /* other routines */
#  define DEBUG_ALL       0xFFFF    /* all debugging data */

#  define DEBUG_LOG(FLAG, P0, P1, P2, P3, P4, P5, P6) \
       if (xtemac_debug & FLAG) logMsg(P0, P1, P2, P3, P4, P5, P6);
#else
#  define DEBUG_LOG(FLAG, P0, P1, P2, P3, P4, P5, P6)
#endif

#define BUF_ALIGNMENT   8     /* 64-bit min buffer alignment */
#define DESC_ALIGNMENT  4     /* 32-bit min buffer descriptor alignment */
#define TX_COPY_BUF_ID  9     /* SGDMA Tx tracking ID type for a locally
                                 managed copy buffer */

/**************************** Type Definitions *******************************/

/* This structure is used when the InitString is parsed and translated at
 * load time
 */
typedef struct
{
    int     Unit;
    Xuint16 DeviceId;
    Xuint8* MacAddrPtr;
    Xuint32 BufBaseAddr;
    Xuint32 BufBytes;
    Xuint32 BdBaseAddr;
    Xuint32 BdBytes;
    Xuint32 RxBdCnt;
    Xuint32 TxBdCnt;
    int     CacheLineSize;
    Xuint32 IntrId;
    Xuint32 MTU;
} InitStringData;


/***************** Macros (Inline Functions) Definitions *********************/

/* Utility macros used on the END_OBJ structure */
#define END_HADDR(EndPtr) \
                ((EndPtr)->mib2Tbl.ifPhysAddress.phyAddress)

#define END_HADDR_LEN(EndPtr) \
                ((EndPtr)->mib2Tbl.ifPhysAddress.addrLength)

/* Shortcut to XTemac instance from End structure */
#define MAC (&DriverPtr->Mac)

/* Rather than use XTemac functions, use DMA directly for BD related access */
#define RXBD_ALLOC(Num, BdPtrPtr)  XDmaV3_SgBdAlloc(&((MAC)->RecvDma), (Num), (BdPtrPtr))
#define RXBD_UNALLOC(Num, BdPtr)   XDmaV3_SgBdUnAlloc(&((MAC)->RecvDma), (Num), (BdPtr))
#define RXBD_TOHW(Num, BdPtr)      XDmaV3_SgBdToHw(&((MAC)->RecvDma), (Num), (BdPtr))
#define RXBD_FROMHW(Num, BdPtrPtr) XDmaV3_SgBdFromHw(&((MAC)->RecvDma), (Num), (BdPtrPtr))
#define RXBD_FREE(Num, BdPtr)      XDmaV3_SgBdFree(&((MAC)->RecvDma), (Num), (BdPtr))
#define RXBD_FREECNT()             XDmaV3_mSgGetFreeCnt(&((MAC)->RecvDma))
#define RXBD_NEXT(BdPtr)           XDmaV3_mSgBdNext(&((MAC)->RecvDma), (BdPtr))
#define RXBD_PREV(BdPtr)           XDmaV3_mSgBdPrev(&((MAC)->RecvDma), (BdPtr))

/* Ack DMA after getting BDs from HW*/
#define RXBD_ACK() \
    {              \
        Xuint32 DmaIsr =                                        \
            XDmaV3_GetInterruptStatus(&((MAC)->RecvDma)) &      \
            XDmaV3_GetInterruptEnable(&((MAC)->RecvDma)) &      \
            (XDMAV3_IPXR_PCTR_MASK | XDMAV3_IPXR_PWBR_MASK);    \
        XDmaV3_SetInterruptStatus(&((MAC)->RecvDma), DmaIsr);   \
    }

#define TXBD_ALLOC(Num, BdPtrPtr)  XDmaV3_SgBdAlloc(&((MAC)->SendDma), (Num), (BdPtrPtr))
#define TXBD_UNALLOC(Num, BdPtr)   XDmaV3_SgBdUnAlloc(&((MAC)->SendDma), (Num), (BdPtr))
#define TXBD_TOHW(Num, BdPtr)      XDmaV3_SgBdToHw(&((MAC)->SendDma), (Num), (BdPtr))
#define TXBD_FROMHW(Num, BdPtrPtr) XDmaV3_SgBdFromHw(&((MAC)->SendDma), (Num), (BdPtrPtr))
#define TXBD_FREE(Num, BdPtr)      XDmaV3_SgBdFree(&((MAC)->SendDma), (Num), (BdPtr))
#define TXBD_FREECNT()             XDmaV3_mSgGetFreeCnt(&((MAC)->SendDma))
#define TXBD_NEXT(BdPtr)           XDmaV3_mSgBdNext(&((MAC)->SendDma), (BdPtr))
#define TXBD_PREV(BdPtr)           XDmaV3_mSgBdPrev(&((MAC)->SendDma), (BdPtr))

/* Ack DMA after getting BDs from HW*/
#define TXBD_ACK() \
    {              \
        Xuint32 DmaIsr =                                        \
            XDmaV3_GetInterruptStatus(&((MAC)->SendDma)) &      \
            XDmaV3_GetInterruptEnable(&((MAC)->SendDma)) &      \
            (XDMAV3_IPXR_PCTR_MASK | XDMAV3_IPXR_PWBR_MASK);    \
        XDmaV3_SetInterruptStatus(&((MAC)->SendDma), DmaIsr);   \
    }


/* TX_COPY_BUF_ALLOC
 *
 * Return the next free TxBuf buffer from the locally managed transmit
 * buffer copy pool. Update TxBuf counters and pointers.
 */
#define TX_COPY_BUF_ALLOC(DrvPtr, BufPtr)                               \
    {                                                                   \
        taskLock();                                                     \
        if (DrvPtr->TxBuf.BufsAvailable == 0)                           \
        {                                                               \
            BufPtr = NULL;                                              \
        }                                                               \
        else                                                            \
        {                                                               \
            BufPtr = DrvPtr->TxBuf.NextBuf;                             \
            DriverPtr->TxBuf.BufsAvailable--;                           \
            DrvPtr->TxBuf.NextBuf += DrvPtr->TxBuf.BufSize;             \
            if (DrvPtr->TxBuf.NextBuf > DriverPtr->TxBuf.HighAddr)      \
            {   /* wrap around */                                       \
                DriverPtr->TxBuf.NextBuf = DriverPtr->TxBuf.BaseAddr;   \
            }                                                           \
        }                                                               \
        taskUnlock();                                                   \
    }
    
/* TX_COPY_BUF_FREE
 *
 * Note the presence of an additional buffer to the locally managed transmit
 * buffer copy pool. Use with TX_COPY_BUF_ALLOC to allocate and return Tx copy
 * buffers.
 */
#define TX_COPY_BUF_FREE(DrvPtr)                \
    {                                           \
        taskLock();                             \
        DrvPtr->TxBuf.BufsAvailable++;          \
        taskUnlock();                           \
    }

/* TX_COPY_BUF_UNALLOC
 *
 * Undo a TX_COPY_BUF_ALLOC operation.
 */
#define TX_COPY_BUF_UNALLOC(DrvPtr)                             \
    {                                                           \
        taskLock();                                             \
        DriverPtr->TxBuf.BufsAvailable++;                       \
        DrvPtr->TxBuf.NextBuf -= DrvPtr->TxBuf.BufSize;         \
        if (DrvPtr->TxBuf.NextBuf < DriverPtr->TxBuf.BaseAddr)  \
        {   /* wrap around */                                   \
            DriverPtr->TxBuf.NextBuf = DriverPtr->TxBuf.HighAddr;       \
        }                                                               \
        taskUnlock();                                                   \
    }
    
/* MIB counters have a new implementation for VxWorks 6.x.
 *
 * We choose to retain the 5.x IN and OUT packet counter to save overhead for 
 * gigabit speeds. If it is desired to use the recommended counter implementation
 * for VxWorks 6.x, change COUNT_IN_PKT and COUNT_OUT_PKT to:
 *
 *     endM2Packet(DriverPtr, (MblkPtr), M2_PACKET_IN)
 *     endM2Packet(DriverPtr, (MblkPtr), M2_PACKET_OUT)
 */
#if (_WRS_VXWORKS_MAJOR >= 6)
#  define COUNT_IN_PKT(MblkPtr)      END_ERR_ADD(&DriverPtr->VxEnd, MIB2_IN_UCAST, 1)
#  define COUNT_OUT_PKT(MblkPtr)     END_ERR_ADD(&DriverPtr->VxEnd, MIB2_OUT_UCAST, 1)
#  define COUNT_IN_DISCARD(MblkPtr)  endM2Packet(&DriverPtr->VxEnd, (MblkPtr), M2_PACKET_IN_DISCARD)
#  define COUNT_OUT_DISCARD(MblkPtr) endM2Packet(&DriverPtr->VxEnd, (MblkPtr), M2_PACKET_OUT_DISCARD)
#  define COUNT_IN_ERROR()           endM2Packet(&DriverPtr->VxEnd, NULL,      M2_PACKET_IN_ERROR)
#  define COUNT_OUT_ERROR(MblkPtr)   endM2Packet(&DriverPtr->VxEnd, (MblkPtr), M2_PACKET_OUT_ERROR)
#else
#  define COUNT_IN_PKT(MblkPtr)      END_ERR_ADD(&DriverPtr->VxEnd, MIB2_IN_UCAST, 1)
#  define COUNT_OUT_PKT(MblkPtr)     END_ERR_ADD(&DriverPtr->VxEnd, MIB2_OUT_UCAST, 1)
#  define COUNT_IN_DISCARD(MblkPtr)
#  define COUNT_OUT_DISCARD(MblkPtr)
#  define COUNT_IN_ERROR()           END_ERR_ADD(&DriverPtr->VxEnd, MIB2_IN_ERRS, 1)
#  define COUNT_OUT_ERROR(MblkPtr)   END_ERR_ADD(&DriverPtr->VxEnd, MIB2_OUT_ERRS, 1)
#endif

/************************** Function Prototypes ******************************/

/* The VxWorks interface functions.  These can be static since they'll be
 * part of a function table passed to VxWorks.
 */
static STATUS Start(XTemacEnd *DriverPtr);
static STATUS Stop(XTemacEnd *DriverPtr);
static STATUS Unload(XTemacEnd *DriverPtr);
static STATUS Ioctl(XTemacEnd *DriverPtr, int Cmd, caddr_t Data);
static STATUS FifoSend(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr);
static STATUS SgSend(XTemacEnd *DriverPtr, M_BLK_ID InMblkPtr);
static STATUS PollSend(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr);
static STATUS PollRecv(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr);
static STATUS MCastAddrGet(END_OBJ *EndPtr, MULTI_TABLE *MultiTablePtr);
static STATUS MCastAddrAdd(END_OBJ *EndPtr, char *AddressPtr);
static STATUS MCastAddrDel(END_OBJ *EndPtr, char *AddressPtr);
static int    MCastAddrFind(XTemacEnd *DriverPtr, char *AddressPtr);

/*
 * Adapter internal functions
 */

/* Handlers called in interrupt context */
static void FifoRecvInt(XTemacEnd *DriverPtr);
static void FifoSendInt(XTemacEnd *DriverPtr, unsigned StatusCnt);
static void SgRecvInt(XTemacEnd *DriverPtr);
static void SgSendInt(XTemacEnd *DriverPtr);
static void ErrorInt(XTemacEnd *DriverPtr, XStatus ErrorClass, Xuint32 Word1,
                     Xuint32 Word2);

/* Handlers called in task context */
static void FifoRecvTask(XTemacEnd *DriverPtr);
static void SgRecvTask(XTemacEnd* DriverPtr);
static void SgSendTask(XTemacEnd* DriverPtr);
static void ErrorTask(XTemacEnd *DriverPtr, XStatus ErrorClass, Xuint32 Word1,
                      Xuint32 Word2);

/* Other internal functions */
static XStatus Mblk2Fifo(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr);
static int IoctlSetDevControl(XTemacEnd *DriverPtr, int Cmd, caddr_t Data);
static void Restart(XTemacEnd *DriverPtr);
static XStatus MemoryInit(XTemacEnd *DriverPtr, InitStringData *InitDataPtr);
static XStatus MemorySgDmaInit(XTemacEnd *DriverPtr, InitStringData *InitDataPtr);
static XStatus ProcessInitString(InitStringData *Dptr, char *InitString);
static STATUS SgSendCoalesce(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr);
static void SgServiceRxRing(XTemacEnd *DriverPtr);
static void SgRecoverRxMblks(XTemacEnd *DriverPtr, int NumBds, XDmaBdV3 *BdPtr);
                            
#ifdef XTEA_CSUM_OFFLOAD
static void CsumTx(M_BLK_ID MblkPtr, XDmaBdV3 *BdPtr);
#endif

/************************** Variable Definitions *****************************/

/*
 * Set this to enable debugging, either at compile time or runtime
 */
#ifdef XMAC_DEBUG
unsigned xtemac_debug = DEBUG_ERROR;
#endif

/* VxWorks function tables. There is a different one for each of the
 * following HW configurations:
 *
 *   - Fifo direct (no DMA)
 *   - Scatter gather DMA (SGDMA)
 */
static NET_FUNCS VxFuncsFifo =
{
    (FUNCPTR)Start,
    (FUNCPTR)Stop,
    (FUNCPTR)Unload,
    (FUNCPTR)Ioctl,
    (FUNCPTR)FifoSend,
    (FUNCPTR)MCastAddrAdd,
    (FUNCPTR)MCastAddrDel,
    (FUNCPTR)MCastAddrGet,
    (FUNCPTR)PollSend,
    (FUNCPTR)PollRecv,
    endEtherAddressForm,
    endEtherPacketDataGet,
    endEtherPacketAddrGet
};

static NET_FUNCS VxFuncsSgDma =
{
    (FUNCPTR)Start,
    (FUNCPTR)Stop,
    (FUNCPTR)Unload,
    (FUNCPTR)Ioctl,
    (FUNCPTR)SgSend,
    (FUNCPTR)MCastAddrAdd,
    (FUNCPTR)MCastAddrDel,
    (FUNCPTR)MCastAddrGet,
    (FUNCPTR)PollSend,
    (FUNCPTR)PollRecv,
    endEtherAddressForm,
    endEtherPacketDataGet,
    endEtherPacketAddrGet
};


/*****************************************************************************/
/**
 * Load an Ethernet device into the VxWorks MUX, and associate a driver with
 * the device. This function is called by the operating system to initialize
 * Ethernet devices. Each call to load a specific device is really a two-pass
 * call. The first pass contains an empty initialization string, and the second
 * pass the real init string. The function is to return the name of the Ethernet
 * device on the first pass. On the second pass, the function parses the init
 * string and initialize everything that needs initializing.
 *
 * In VxWorks, the kernel has been initialized by the time this load function
 * is called. So allocating/using kernel resources such as semaphores and
 * dynamic memory management (e.g., malloc) is OK.
 *
 * @param InitString is the initialization string specific to the Xilinx
 *        Ethernet device. It is typically created in the BSP based on system
 *        parameters. See the comments near the beginning of xtemac_end-
 *        _adapter.c for details on its format. On the first call to this
 *        function, InitString is an output parameter that gets the name of the
 *        device.
 *
 * @param BspPtr is BSP-specific information. This is currently unused.
 *
 * @return The return value is a pointer to the END_OBJ structure. On the first
 *         call to this function, the return value is NULL and the InitString
 *         parameter gets the name of the device. On error, the return value is
 *         NULL.
 *
 * @note
 *
 * This function is VxWorks-specific. It plugs into the endLoad interface.
 *
 ******************************************************************************/
END_OBJ *XTemacEndAdapter_Load(char *InitString, void *BspPtr)
{
    XTemacEnd *DriverPtr;
    NET_FUNCS *FuncSet;
    InitStringData InitData;
    XStatus Status;
    Xuint16 Speed;
    Xuint32 DevFlags;
    
    DEBUG_LOG(DEBUG_MISC, "Load called\n", 1, 2, 3, 4, 5, 6);
    
    /* In case VxWorks sends us a null init string */
    if (InitString == NULL)
    {
        return(NULL);
    }

    /* The first time this function is called, VxWorks passes an empty string
     * and expects the name of the driver in return.
     */
    if (InitString[0] == 0)
    {
        strcpy(InitString, XTE_DEVICE_NAME);
        return((END_OBJ *)NULL);
    }

    /* Allocate the device structure */
    DriverPtr = (XTemacEnd*)calloc(sizeof(XTemacEnd), 1);
    if (DriverPtr == NULL)
    {
        DEBUG_LOG(DEBUG_ERROR, "Unable to calloc DriverPtr\n", 1, 2, 3, 4, 5, 6);
        return(NULL);
    }

    /* Process the init string */
    if (ProcessInitString(&InitData, InitString) != XST_SUCCESS)
    {
        /* failures are DEBUG_LOG'd in ProcessInitString */
        goto ErrorExit;
    }

    /* Copy selected init data into driver structure */
    DriverPtr->VxUnit = InitData.Unit;
    DriverPtr->IntrId = InitData.IntrId;
    DriverPtr->MaxFrameSz = InitData.MTU + XTE_HDR_SIZE + XTE_TRL_SIZE;    
    macptr = MAC;
    
    /* Initialize the level 1 driver */
    Status = XTemac_Initialize(MAC, InitData.DeviceId);
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "Error initializing mac, code %d\n",
                  Status, 2, 3, 4, 5, 6);
        goto ErrorExit;
    }
    
    /* Set the device's MAC address */
    Status = XTemac_SetMacAddress(MAC, InitData.MacAddrPtr);
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "Error setting mac address, code %d\n",
                  Status, 2, 3, 4, 5, 6);
        goto ErrorExit;
    }


    /* Device options to disable */
    DevFlags = XTE_REPORT_RXERR_OPTION;   /* Don't report dropped frames */    
    XTemac_ClearOptions(MAC, DevFlags);
    
    /* Device options to enable */
    DevFlags = XTE_MULTICAST_CAM_OPTION;
    
#ifdef XTEA_CSUM_OFFLOAD
        DriverPtr->HwCaps.cap_available |= IFCAP_JUMBO_MTU;
#endif
        
    if (InitData.MTU > XTE_MTU)
    {
        DevFlags |= XTE_JUMBO_OPTION;
    }

    XTemac_SetOptions(MAC, DevFlags);

    /* Determine the method of frame Rx/Tx based on device capabilities
     *   - Assign function table
     *   - Set interrupt handler callbacks
     *   - Set which events will cause interrupts. Actual interrupt will not
     *     occur until XTemac_Start() and intEnable() is called.
     */
    if (XTemac_mIsSgDma(MAC))
    {
        FuncSet = &VxFuncsSgDma;
        XTemac_SetHandler(MAC, XTE_HANDLER_SGSEND, SgSendInt, DriverPtr);
        XTemac_SetHandler(MAC, XTE_HANDLER_SGRECV, SgRecvInt, DriverPtr);

        /* Configure the scatter-gather DMA packet threshold and packet wait
         * bound to default values for both send and receive directions.
         */
        Status  = XTemac_IntrSgCoalSet(MAC, XTE_SEND, 1, 0);
        Status |= XTemac_IntrSgCoalSet(MAC, XTE_RECV, 1, 0);

        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR,
                      "Error setting up interrupt coalescing, code %d\n",
                      Status, 2,3,4,5,6);
            goto ErrorExit;
        }

        XTemac_IntrSgEnable(MAC, XTE_SEND | XTE_RECV);

#ifdef XTEA_CSUM_OFFLOAD
        /* Checksum offloading capabilities */
        if (XTemac_mIsRxCsum(MAC))
        {
            DriverPtr->HwCaps.cap_available |= IFCAP_RXCSUM;
            DriverPtr->HwCaps.csum_flags_rx |= (CSUM_IP | CSUM_UDP | CSUM_TCP);
        }
        
        if (XTemac_mIsTxCsum(MAC))
        {
            DriverPtr->HwCaps.cap_available |= IFCAP_TXCSUM;
            DriverPtr->HwCaps.csum_flags_tx |= (CSUM_UDP | CSUM_TCP);
        }

        DriverPtr->HwCaps.cap_enabled = DriverPtr->HwCaps.cap_available;
#endif    
    }
    else  /* FIFO direct mode */
    {
        FuncSet = &VxFuncsFifo;
        XTemac_SetHandler(MAC, XTE_HANDLER_FIFORECV, FifoRecvInt, DriverPtr);
        XTemac_SetHandler(MAC, XTE_HANDLER_FIFOSEND, FifoSendInt, DriverPtr);
        XTemac_IntrFifoEnable(MAC, XTE_SEND | XTE_RECV);

        /* Don't enable transmit status overrun interrupts for this mode */
        XTemac_ClearOptions(MAC, XTE_REPORT_TXSTATUS_OVERRUN_OPTION);
    }
    
    /* Set the error handler callback */
    XTemac_SetHandler(MAC, XTE_HANDLER_ERROR, ErrorInt, DriverPtr);

    /* Initialize the VxWorks END_OBJ structure */
    if (END_OBJ_INIT(&DriverPtr->VxEnd,
                     (DEV_OBJ *)DriverPtr,
                     XTE_DEVICE_NAME,
                     DriverPtr->VxUnit,
                     FuncSet,
                     XTE_DEVICE_DESC) == ERROR)
    {
        DEBUG_LOG(DEBUG_ERROR,"END_OBJ_INIT failed\n", 1,2,3,4,5,6);
        goto ErrorExit;
    }

    /* Here is a problem area. When running 10/100 MII or 10/100/1000 Tri Mode
     * GMII, the link speed must be explicitly set in the MAC core to lock in 
     * the clock from the PHY. If 1000BaseX mode is enabled, then we know to set
     * for 1000 Mbps. Otherwise we must query the PHY.
     *
     * To query the PHY, we must know first what kind of PHY it is as the actual
     * negotiated speed is not part of the MII standard register set. Worse yet,
     * we don't know even if the link is up. As a result, 1000 Mbps is used as
     * a default. If the real speed turns out to be 10 or 100 Mbps, then the MAC
     * will not pass traffic.
     *
     * This is an ugly HW limitation that forces us to know which PHY we are on.
     * The most robust solution would be to utilize a unique PHY driver, using
     * the auto negotiate interrupt built into the TEMAC. When the interrupt
     * occurs, query the PHY and adjust the clock
     */
    if (XTemac_mGetPhysicalInterface(MAC) == XTE_PHY_TYPE_MII)
    {
        Speed = 100;
    }
    else
    {
        Speed = 1000;
    }

    XTemac_SetOperatingSpeed(MAC, Speed);
    printf("temac: warning - MII clock is defaulted to %d Mbps\n", Speed);
    taskDelay(sysClkRateGet()/2);

    /* Initialize the MIB element of the END_OBJ structure, which informs
     * VxWorks of several specifics regarding the device interface (type,
     * MTU, address).
     */
    if (END_MIB_INIT(&DriverPtr->VxEnd,
                     M2_ifType_ethernet_csmacd,
                     InitData.MacAddrPtr,
                     XTE_MAC_ADDR_SIZE,
                     InitData.MTU,
                     1000) == ERROR)
    {
        DEBUG_LOG(DEBUG_ERROR,"END_OBJ_INIT failed\n", 1,2,3,4,5,6);
        goto ErrorExit;
    }

    /* Allocate and setup buffer memory space. If configured for SG DMA, then
     * buffer descriptor space is setup as well.
     */
    if (MemoryInit(DriverPtr, &InitData) != XST_SUCCESS)
    {
        goto ErrorExit;  /* failures are DEBUG_LOG'd in MemoryInit */
    }

    /* Create a mutual exclusion semaphore to guard configuration of the
     * driver between Start, Stop, and Option setting functions. 
     */
    DriverPtr->SemConfig = semBCreate(SEM_Q_FIFO, SEM_FULL);
    if (DriverPtr->SemConfig == NULL)
    {
        DEBUG_LOG(DEBUG_ERROR,"Semaphore creation failed\n", 1,2,3,4,5,6);
        goto ErrorExit;
    }

    /* Set the flags in the END_OBJ to indicate the device has been
     * initialized and is ready to be started. 
     */
    if (END_OBJ_READY(&DriverPtr->VxEnd, IFF_MULTICAST | IFF_NOTRAILERS |
                      IFF_BROADCAST | IFF_SIMPLEX) == ERROR)
    {
        DEBUG_LOG(DEBUG_ERROR,"END_OBJ_READY failed\n", 1,2,3,4,5,6);
        goto ErrorExit;
    }

    /* Success! */
    return(&DriverPtr->VxEnd);
    
    /* error encountered during initialization */
ErrorExit:
    free((char*)DriverPtr);
    return(NULL);
}


/*****************************************************************************/
/**
 * Get a pointer to the L1 driver used for the given unit number. 
 *
 * @param UnitNumber The unit number of the device assigned at load time
 *
 * @return
 *
 * A pointer to the L1 driver instance used for this unit number, or NULL if the
 * device has not been loaded into the Mux.
 *
 ******************************************************************************/
XTemac *XTemacEndAdapter_GetDriver(int UnitNumber)
{
    XTemacEnd *End;

    End = (XTemacEnd*)endFindByName(XTE_DEVICE_NAME, UnitNumber);
    if (End == NULL)
    {
        return(NULL);
    }
    else
    {
        return(&End->Mac);
    }
}


/******************************************************************************
*
* FUNCTION:
*
* Unload
*
* DESCRIPTION:
*
* Unload the driver from the MUX interface and free any dynamically allocated
* memory. The device is stopped even though the MUX should have done it.
* device.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* RETURN VALUE:
*
* OK (expects to be successful)
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS Unload(XTemacEnd *DriverPtr)
{
    XASSERT_NONVOID(DriverPtr != NULL);

    DEBUG_LOG(DEBUG_MISC, "Unload: called\n",1,2,3,4,5,6);

    /* Stop the device just to be sure */
    XTemac_Stop(MAC);

    /* Call required END macro to unload the device */
    END_OBJECT_UNLOAD(&DriverPtr->VxEnd);

    /* Free up allocated resources */
    if (DriverPtr->SgDma.IsBdMalloc)
    {
        cacheDmaFree((void*)DriverPtr->SgDma.BdSpaceAddr);
    }

    if (DriverPtr->IsBufMalloc)
    {
        cacheDmaFree((void*)DriverPtr->BufSpaceAddr);
    }

    if (DriverPtr->RxBuf.ClConfig.memArea != NULL)
    {
        free(DriverPtr->RxBuf.ClConfig.memArea);
    }

    semDelete(DriverPtr->SemConfig);

    return(OK);
}

/******************************************************************************
*
* FUNCTION:
*
* Start
*
* DESCRIPTION:
*
* Starts the device referred to by the END_OBJ structure. This is part of the
* VxWorks function table and is typically called by muxDevStart().
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* RETURN VALUE:
*
* OK on success, ERROR if VxWorks has not yet set a receive callback function
* or if the driver cannot be started.
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS Start(XTemacEnd *DriverPtr)
{
    XStatus Result;
    Xuint32 Options;
    XInterruptHandler Isr;

    XASSERT_NONVOID(DriverPtr != NULL);

    DEBUG_LOG(DEBUG_MISC, "Start: called\n",1,2,3,4,5,6);
    
    /* Verify that the MUX set the receive handler */
    if (DriverPtr->VxEnd.receiveRtn == XNULL)
    {
        DEBUG_LOG(DEBUG_ERROR, "Start: error, no receiveRtn\n",1,2,3,4,5,6);
        return(ERROR);
    }

    /* Take the config semaphore for mutex between Start/Stop */
    semTake(DriverPtr->SemConfig, WAIT_FOREVER);

    /* Just to be safe, stop the device first */
    XTemac_Stop(MAC);

    /* If the device is not configured for polled mode, connect to the interrupt
     * controller and enable interrupts. Pick the right interrupt handler for
     * the driver depending on how it's configured.
     */
    Options = XTemac_GetOptions(MAC);
    if (!(Options & XTE_POLLED_OPTION))
    {
        if (XTemac_mIsSgDma(MAC))
        {
            Isr = XTemac_IntrSgHandler;        /* configured for SGDMA */
        }
        else
        {
            Isr = XTemac_IntrFifoHandler;      /* using FIFOs directly */
        }

        intConnect(INUM_TO_IVEC(DriverPtr->IntrId), (VOIDFUNCPTR)Isr, (int)MAC);
        intEnable(DriverPtr->IntrId);
    }

    Result = XTemac_Start(MAC);
    semGive(DriverPtr->SemConfig);

    if (Result != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "Start: error, XTemac_Start returned %d\n",
                   Result,2,3,4,5,6);
        return(ERROR);
    }
    
    END_FLAGS_SET (&DriverPtr->VxEnd, IFF_UP | IFF_RUNNING);
    
    return(OK);
}

/******************************************************************************
*
* FUNCTION:
*
* Stop
*
* DESCRIPTION:
*
* Stops the device referred to by the END_OBJ structure.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* RETURN VALUE:
*
* OK on success, ERROR if the device cannot be stopped.
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS Stop(XTemacEnd *DriverPtr)
{
    XASSERT_NONVOID(DriverPtr != NULL);

    DEBUG_LOG(DEBUG_MISC, "Stop: called\n",1,2,3,4,5,6);
    
    /* Take the config semaphore for mutex between Stop/Start/SetOptions */
    semTake(DriverPtr->SemConfig, WAIT_FOREVER);
    XTemac_Stop(MAC);
    semGive(DriverPtr->SemConfig);

    END_FLAGS_CLR(&DriverPtr->VxEnd, IFF_UP | IFF_RUNNING);

    return(OK);
}


/*****************************************************************************/
/*****************************************************************************
 * Polled mode related functions. 
 *****************************************************************************/
/*****************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* PollSend
*
* DESCRIPTION:
*
* Send an Ethernet frame to the driver in polled mode. The Ethernet frame is
* bundled inside a VxWorks Mblk chain. It is assumed that protocol stack
* supplies a correctly formatted Ethernet frame, including the destination and
* source addresses, the type/length field, and the data field. It is also
* assumed that VxWorks does not pad short frames nor append FCS at the end of
* the frame. 
*
* The driver returns an error if it has not been configured for polled mode.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk chain containing the frame to be sent. This Mblk
* chain is not freed upon return from the function (as is done in non-polled
* mode).
*
* RETURN VALUE:
*
*   OK              The frame was successfully transmitted
*   ERROR           The driver's PollSend function failed.
*
* NOTES:
*
* This function does not free the Mblk passed in after transmission or error.
* It's not clear whether VxWorks expects this or not.  Other drivers (Lance and
* Walnut) did not free it, so neither do we.
*
******************************************************************************/
static STATUS PollSend(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    XStatus Status;
    Xuint32 TxStatus;
    M_BLK_ID MblkCurPtr = MblkPtr;

    /* Will this packet fit in the FIFO */
    if (XTemac_FifoGetFreeBytes(MAC, XTE_SEND) < MblkPtr->mBlkPktHdr.len)
    {
        DEBUG_LOG(DEBUG_ERROR, "PollSend: No room in FIFO\n",1,2,3,4,5,6);
        COUNT_OUT_ERROR(MblkPtr);
        return(ERROR);
    }

    /* Loop through all Mblks for this packet */
    do
    {
        if (MblkCurPtr->mBlkHdr.mNext == NULL)
        {
            /* Reached end of the packet, finish copy for this Mblk chain */
            Status = XTemac_FifoWrite(MAC,
                                      MblkCurPtr->mBlkHdr.mData,
                                      MblkCurPtr->mBlkHdr.mLen,
                                      XTE_END_OF_PACKET);

            /* Initiate transmit */
            if (Status == XST_SUCCESS)
            {
                Status = XTemac_FifoSend(MAC, MblkPtr->mBlkPktHdr.len);
            }
            
            break;
        }
        else
        {
            /* More data in this Mblk chain, copy it to the FIFO */
            Status = XTemac_FifoWrite(MAC,
                                      MblkCurPtr->mBlkHdr.mData,
                                      MblkCurPtr->mBlkHdr.mLen,
                                      XTE_PARTIAL_PACKET);
            MblkCurPtr = MblkCurPtr->mBlkHdr.mNext;
        }
    } while(Status == XST_SUCCESS);

    /* Any errors getting data into the FIFO? */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "PollSend: API error code %d\n",
                  Status,2,3,4,5,6);
        COUNT_OUT_ERROR(MblkPtr);
        ErrorTask(DriverPtr, XST_FIFO_ERROR, 0, 0);
        return(ERROR);
    }

    /* Wait for outcome */
    do
    {
        Status = XTemac_FifoQuerySendStatus(MAC, &TxStatus);
    } while (Status == XST_NO_DATA);

    /* Get final status */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "PollSend: Query error code %d, TxStatus %d\n",
                  Status, TxStatus, 3,4,5,6);
        COUNT_OUT_ERROR(MblkPtr);
        return(ERROR);
    }
        
    /* Success! */
    DEBUG_LOG(DEBUG_TX, "PollSend: %d byte packet sent\n",
              MblkPtr->mBlkPktHdr.len,2,3,4,5,6);
    COUNT_OUT_PKT(MblkPtr);
    return(OK);    
}

/******************************************************************************
*
* FUNCTION:
*
* PollRecv
*
* DESCRIPTION:
*
* Receive an Ethernet frame in polled mode. The Ethernet frame will be placed
* inside a VxWorks Mblk chain. The driver returns an error if it has not been
* configured for polled mode.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk chain containing the buffer into which the
* incoming frame will be stored. We assume the buffer is big enough for the
* MTU as configured.
*
* RETURN VALUE:
*
*   OK              The frame was successfully transmitted
*   ERROR           The driver's PollRecv function failed
*   EAGAIN          No data is available
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS PollRecv(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    XStatus Status;
    Xuint32 FrameLen;

    /* Ask for a status from the device */
    Status = XTemac_FifoQueryRecvStatus(MAC);

    /* Nothing there */
    if (Status == XST_NO_DATA)
    {
        return(EAGAIN);
    }

    /* A state of error is being reported */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "PollRecv: Query error code %d\n",
                  Status, 2,3,4,5,6);
        COUNT_IN_ERROR();
        return(ERROR);
    }

    /* If control reaches this point, there is a frame waiting to be received.
     * Get its length and transfer the FIFO data into the provided buffer.
     */
    Status = XTemac_FifoRecv(MAC, &FrameLen);
    
    if (Status != XST_SUCCESS)
    {
        /* There is a receive indication but no length */
        DEBUG_LOG(DEBUG_ERROR, "PollRecv: FifoRecv error code %d\n",
                  Status, 2,3,4,5,6);
        COUNT_IN_ERROR();
        ErrorTask(DriverPtr, XST_FIFO_ERROR, 0, 0);
        return(ERROR);
    }

    Status = XTemac_FifoRead(MAC, MblkPtr->mBlkHdr.mData, FrameLen,
                             XTE_END_OF_PACKET);

    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "PollRecv FifoRead error code %d\n",
                  Status,2,3,4,5,6);
        COUNT_IN_ERROR();
        ErrorTask(DriverPtr, XST_FIFO_ERROR, 0, 0);
        return(ERROR);
    }

    /* Finish up the packet and exit */
    MblkPtr->mBlkHdr.mFlags |= M_PKTHDR;
    MblkPtr->mBlkHdr.mLen = FrameLen;
    MblkPtr->mBlkPktHdr.len = FrameLen;

    DEBUG_LOG(DEBUG_RX, "PollRecv: %d byte packet received\n",
              FrameLen,2,3,4,5,6);
    COUNT_IN_PKT(MblkPtr);
    return(OK);
}

/*****************************************************************************/
/*****************************************************************************
 * FIFO Direct related transmit and receive functions and handlers. 
 *****************************************************************************/
/*****************************************************************************/


/******************************************************************************
*
* FUNCTION:
*
* FifoSend
*
* DESCRIPTION:
*
* Send an Ethernet frame to the driver. The Ethernet frame is bundled inside a
* VxWorks Mblk chain.  It is assumed that protocol stack supplies a correctly
* formatted Ethernet frame, including the destination and source addresses, the
* type/length field, and the data field.  It is also assumed that VxWorks does
* not pad short frames nor append FCS at the end of the frame.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk chain containing the frame to be sent
*
* RETURN VALUE:
*
*   OK              The frame was sent successfully.
*   END_ERR_BLOCK   Can't send due to insufficient resources in the HW.
*   ERROR           A level 1 driver API function returned an error.
*
* NOTES:
*
* We are assuming that VxWorks sends down only one packet at a time.
*
******************************************************************************/
static STATUS FifoSend(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    XStatus Status;

    END_TX_SEM_TAKE(&DriverPtr->VxEnd, WAIT_FOREVER);

    /* Will this packet fit in the FIFO */
    if (XTemac_FifoGetFreeBytes(MAC, XTE_SEND) < MblkPtr->mBlkPktHdr.len)
    {
        /* Out of resources. The return code causes the transmit channel to
         * block until restarted.
         */
        DriverPtr->TxIsBlocked = XTRUE;        
        DEBUG_LOG(DEBUG_ERROR, "FifoSend: resource blocked\n", 1,2,3,4,5,6);
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(END_ERR_BLOCK);
    }

    Status = Mblk2Fifo(DriverPtr, MblkPtr);
    

    /* Had a problem */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "FifoSend: API error, code %d\n",
                  Status,2,3,4,5,6);
        COUNT_OUT_ERROR(MblkPtr);
        netMblkClChainFree(MblkPtr);
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(ERROR);
    }

    /* Packet is on its way */
    DEBUG_LOG(DEBUG_TX, "FifoSend: %d byte packet sent\n",
              MblkPtr->mBlkPktHdr.len,2,3,4,5,6);
    COUNT_OUT_PKT(MblkPtr);
    netMblkClChainFree(MblkPtr);
    END_TX_SEM_GIVE(&DriverPtr->VxEnd);
    return(OK);    
}


/******************************************************************************
*
* FUNCTION:
*
* FifoSendInt
*
* DESCRIPTION:
*
* Callback function (called from level 1 driver) to handle confirmation of
* transmit events when in FIFO direct mode. This function is expected to
* execute in interrupt context.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
* StatusCnt is the number of successfull transmit statuses received
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void FifoSendInt(XTemacEnd *DriverPtr, unsigned StatusCnt)
{
    DEBUG_LOG(DEBUG_TX, "FifoSendInt: called, StatusCnt=%d\n", StatusCnt,
              2,3,4,5,6);
    
    /* Restart the transmit channel if it is resource blocked */
    if (DriverPtr->TxIsBlocked)
    {
        DriverPtr->TxIsBlocked = XFALSE;        
        DEBUG_LOG(DEBUG_TX, "FifoSendInt: unblocked tx channel\n", 1,2,3,4,5,6);
        if (netJobAdd((FUNCPTR)muxTxRestart, (int)&DriverPtr->VxEnd,
                      0, 0, 0, 0) != OK)
        {
            DEBUG_LOG(DEBUG_ERROR,"FifoSendInt: netJobAdd failed, errno=%d\n",
                      errno,2,3,4,5,6);
        }
    }
}


/******************************************************************************
*
* FUNCTION:
*
* FifoRecvInt
*
* DESCRIPTION:
*
* ISR level handler for FIFO direct receive channel.
*
* Does minimal work. Notifies the net task that a new frame is ready to be
* transferred from the HW to the stack.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void FifoRecvInt(XTemacEnd *DriverPtr)
{
    /* Disable further receive interrupts or else we wind up in an exception
     * loop
     */
    XTemac_IntrFifoDisable(MAC, XTE_RECV);

    DEBUG_LOG(DEBUG_RX, "FifoRecvInt: called\n", 1,2,3,4,5,6);
    
    /* Transfer of data from driver to stack occurs in task context */
    if (netJobAdd((FUNCPTR)FifoRecvTask, (int)DriverPtr, 0, 0, 0, 0) != 0)
    {        
        DEBUG_LOG(DEBUG_ERROR,"FifoRecvInt: netJobAdd failed, errno=%d\n",
                  errno,2,3,4,5,6);
    }
}

/******************************************************************************
*
* FUNCTION:
*
* FifoRecvTask
*
* DESCRIPTION:
*
* Task level handler for FIFO direct receive channel.
*
* Transfers data from the receive packet FIFO to the stack. All frames that
* are present in the FIFO are processed.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void FifoRecvTask(XTemacEnd *DriverPtr)
{
    M_BLK_ID MblkPtr;
    XStatus Status;
    Xuint32 FrameLen;

    /* While there are packets left to get */
    while(1)
    {
        if (!XTemac_mIsRecvFrame(MAC))
        {
            break;  /* No more frames to get */
        }

        /* Get a new buffer to fill with the received frame */
        MblkPtr = netTupleGet(&DriverPtr->RxBuf.NetPool,
                              DriverPtr->RxBuf.ClSize,
                              M_DONTWAIT, MT_DATA, FALSE);
        
        if (MblkPtr == NULL)
        {
            DEBUG_LOG(DEBUG_ERROR, "FifoRecvTask: Out of buffers\n",
                      1,2,3,4,5,6);
            COUNT_IN_ERROR();
            DriverPtr->VxEndLastError.pMesg = "Rx out of buffers";
#ifdef END_ERR_NO_BUF
            DriverPtr->VxEndLastError.errCode = END_ERR_NO_BUF;
#else
            DriverPtr->VxEndLastError.errCode = END_ERR_WARN;
#endif
            muxError(DriverPtr, &DriverPtr->VxEndLastError);
            break;
        }

        /* First get length of received packet. Since we passed the
         * XTemac_mIsRecvFrame test above, we know a packet is waiting so any
         * error here is bad. Same story if there is an error reading the frame
         * out of the packet FIFO.
         */
        Status = XTemac_FifoRecv(MAC, &FrameLen);
        if (Status == XST_SUCCESS)
        {
            /* Copy frame from device to buffer */
            Status = XTemac_FifoRead(MAC, MblkPtr->mBlkHdr.mData, FrameLen,
                                     XTE_END_OF_PACKET);
        }

        /* If an error, then the packet FIFO is likely out of whack with the
         * length FIFO. A reset is needed.
         */
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR, "FifoRecvTask: FifoRecv error %d\n",
                      Status, 2,3,4,5,6);
            
            netMblkClChainFree(MblkPtr);
            COUNT_IN_ERROR();
            break;
        }

        /* Set packet flags and length then send up stack */
        MblkPtr->mBlkHdr.mLen = FrameLen;
        MblkPtr->mBlkPktHdr.len = FrameLen;
        MblkPtr->mBlkHdr.mFlags |= M_PKTHDR;
        
        COUNT_IN_PKT(MblkPtr);
        END_RCV_RTN_CALL(&DriverPtr->VxEnd, MblkPtr);
        
        DEBUG_LOG(DEBUG_RX, "FifoRecvTask: %d byte packet passed to stack\n",
                  FrameLen, 2,3,4,5,6);
    }

    /* Re-enable receive interrupts */
    XTemac_IntrFifoEnable(MAC, XTE_RECV);
}



/*****************************************************************************/
/*****************************************************************************
 * SGDMA related transmit and receive functions and handlers. 
 *****************************************************************************/
/*****************************************************************************/

/******************************************************************************
*
* FUNCTION:
*
* SgSend
*
* DESCRIPTION:
*
* Send an Ethernet frame to the driver when the device is configured for
* SGDMA. The Ethernet frame is bundled inside a VxWorks Mblk chain.
* It is assumed that protocol stack supplies a correctly formatted Ethernet
* frame, including the destination and source addresses, the type/length field,
* and the data field.  It is also assumed that VxWorks does not pad short frames
* nor append FCS at the end of the frame.
*
* If alignment requirments are not met, then a copy of the frame is made before
* submitting it to HW
*
* If the Mblk chain has too many fragments, they are coalesced into a single
* buffer.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk chain containing the frame to be sent
*
* RETURN VALUE:
*
*   OK              The frame was successfully queued to the DMA engine
*   END_ERR_BLOCK   Insufficient resources. Out of buffer descriptors or out of
*                   copy buffers.
*   ERROR           One of the Level 1 driver's API functions failed. The incoming
*                   Mblk chain is freed when this error is returned.
*
******************************************************************************/
static STATUS SgSend(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    XDmaBdV3 *Bd1stPtr, *BdCurPtr;
    M_BLK_ID MblkCurPtr;
    Xuint32 BdAllocCnt, BdsLeft;
    XStatus Status;

    /* Gain access to the send channel */
    END_TX_SEM_TAKE(&DriverPtr->VxEnd, WAIT_FOREVER);

    /* Must be at least 1 BD available */
    BdAllocCnt = TXBD_FREECNT();
    if (BdAllocCnt == 0)
    {
        DEBUG_LOG(DEBUG_TX, "SgSend: out of descriptors\n", 1,2,3,4,5,6);
        DriverPtr->TxIsBlocked = XTRUE;
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(END_ERR_BLOCK);
    }

    /* If DMA engine cannot tranfer on byte alignment, must use different
     * send algorithm that coalesces Mblk chain into a single aligned buffer */
    if (!XTemac_mIsTxDre(MAC))
    {
        return(SgSendCoalesce(DriverPtr, MblkPtr));
    }
    
    /* Allocate BD(s) from the transmit channel */
    BdAllocCnt = min(BdAllocCnt, XTEA_SG_MAX_FRAGS);
    
    taskLock();
    Status = TXBD_ALLOC(BdAllocCnt, &Bd1stPtr);
    taskUnlock();
    
    /* Since we already know there should be at least this many BDs, a failure
     * here means something very bad happened */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "SgSend: error allocating BDs, code %d\n",
                  Status, 2,3,4,5,6);
        goto ReturnFatal;
    }
    
    /* Loop until we've either exausted the number of BDs allocated or until
     * we've processed all mblks */
    MblkCurPtr = MblkPtr;
    BdCurPtr = Bd1stPtr;
    BdsLeft = BdAllocCnt;
    
    while (BdsLeft)
    {
        /* Filter out any 0 length mblks */
        if (MblkCurPtr->mBlkHdr.mLen == 0)
        {
            MblkCurPtr = MblkCurPtr->mBlkHdr.mNext;
            if (MblkCurPtr == NULL)
            {
                /* The mblk chain ends with an 0 length mblk. Make sure we didn't
                 * receive a chain of empty mblks. If we did then just return
                 * because there is nothing to send. If there was data, then
                 * mark the last BD that contains actual data */
                if (BdsLeft != BdAllocCnt)
                {
                    XDmaBdV3_mSetLast(TXBD_PREV(BdCurPtr));
                    break;  /* exit while (BdsLeft) */
                }
                else
                {
                    taskLock();
                    Status = TXBD_UNALLOC(BdAllocCnt, Bd1stPtr);
                    taskUnlock();
                    netMblkClChainFree(MblkPtr);
                    END_TX_SEM_GIVE(&DriverPtr->VxEnd);
                    return(OK);
                }
            }
            continue;
        }
        
        /* Flush the cache for this mBlk */
        CACHE_USER_FLUSH(MblkCurPtr->mBlkHdr.mData, MblkCurPtr->mBlkHdr.mLen);

        /* Assign source address of buffer data, length, and tracking ID */
        XDmaBdV3_mSetLength(BdCurPtr, MblkCurPtr->mBlkHdr.mLen);
        XDmaBdV3_mSetBufAddrLow(BdCurPtr,
                                CACHE_DMA_VIRT_TO_PHYS(MblkCurPtr->mBlkHdr.mData));

        /* Use the 1st mblk as the tracking ID for every BD in this chain. When
         * SgSendTask cleans up afterwards, it will call netMblkClChainFree once
         * to free all the mblks in this chain.
         *
         * We need to set the id to something because SgSendTask needs to tell
         * the difference between an mblk or a TX copy buffer if SgSendCoalesce
         * was used to setup other outbound packet buffers
         */
        XDmaBdV3_mSetId(BdCurPtr, MblkPtr);
        BdsLeft--;

        /* Move to the next mblk */
        MblkCurPtr = MblkCurPtr->mBlkHdr.mNext;
        if (MblkCurPtr == NULL)
        {
            XDmaBdV3_mSetLast(BdCurPtr);      /* We are done */
            BdCurPtr = TXBD_NEXT(BdCurPtr);
            break;
        }
        else
        {
            XDmaBdV3_mClearLast(BdCurPtr);    /* More mblks to process */
            BdCurPtr = TXBD_NEXT(BdCurPtr);
        }
    } 

    /* If we have run out of BDs before Mblks, then undo everything and coalesce
     * mblks into one buffer then send that buffer */
    if (MblkCurPtr != NULL)
    {
        taskLock();
        Status = TXBD_UNALLOC(BdAllocCnt, Bd1stPtr);
        taskUnlock();

        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR,
                      "SgSend: error unallocating all BDs, code %d\n",
                      Status, 2,3,4,5,6);
            goto ReturnFatal;
        }
        
        return(SgSendCoalesce(DriverPtr, MblkPtr));
    }

    /* Unallocate unused BDs (BdCurPtr should point to 1st unused BD) */
    else if (BdsLeft)
    {
        taskLock();
        Status = TXBD_UNALLOC(BdsLeft, BdCurPtr);
        taskUnlock();
        
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR,
                      "SgSend: error unallocating unused BDs, code %d\n",
                      Status, 2,3,4,5,6);
            goto ReturnFatal;
        }
    }

    /* Do CSUM offload setup on 1st BD in chain */
#ifdef XTEA_CSUM_OFFLOAD
    if (DriverPtr->HwCaps.cap_enabled & IFCAP_TXCSUM)
    {
        CsumTx(MblkPtr, Bd1stPtr);
    }
    else
    {
        XTemac_mSgSendBdCsumDisable(Bd1stPtr);
    }
#endif
    
    /* Commit BDs to HW */
    taskLock();
    Status = TXBD_TOHW(BdAllocCnt - BdsLeft, Bd1stPtr);
    taskUnlock();

    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR,
                  "SgSend: error committing BDs, code %d\n",
                  Status, 2,3,4,5,6);
        goto ReturnFatal;
    }
    
    /* Success */
    COUNT_OUT_PKT(MblkPtr);
    END_TX_SEM_GIVE(&DriverPtr->VxEnd);
    return(OK);

    /* Something has gone wrong with the descriptor ring. Device must go down, no
     * other way to recover other than unload + reload */
    ReturnFatal:
        XTemac_Stop(MAC);
        COUNT_OUT_ERROR(MblkPtr);
        netMblkClChainFree(MblkPtr);
        DriverPtr->VxEndLastError.errCode = END_ERR_DOWN;
        DriverPtr->VxEndLastError.pMesg = "Fatal SGDMA Tx error";
        muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(ERROR);
}


/******************************************************************************
*
* FUNCTION:
*
* SgSendCoalesce
*
* DESCRIPTION:
*
* This function is called by SgSend when conditions warrant coalescing all
* mblks in a Tx chain to a single buffer.
*
* A locally managed Tx buffer is allocated and enqueued to the SGDMA engine.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk chain containing the frame to be sent
*
* RETURN VALUE:
*
*   OK              The frame was successfully queued to the DMA engine
*   END_ERR_BLOCK   Insufficient resources. Out of buffer descriptors or out of
*                   copy buffers.
*   ERROR           One of the Level 1 driver's API functions failed. The incoming
*                   Mblk chain is freed when this error is returned.
*
******************************************************************************/
static STATUS SgSendCoalesce(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    XStatus Status;
    XDmaBdV3 *BdPtr;
    Xuint8 *DataPtr;
    int PktLen;
    
    /* Allocate one of our copy buffers */
    TX_COPY_BUF_ALLOC(DriverPtr, DataPtr);
    
    if (DataPtr == NULL)
    {
        DEBUG_LOG(DEBUG_TX, "SgSendCoalesce: out of copy buffers\n", 1,2,3,4,5,6);
        DriverPtr->TxIsBlocked = XTRUE;
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(END_ERR_BLOCK);
    }

    /* Copy from mblk to buffer discard the mblk */
    PktLen = netMblkToBufCopy(MblkPtr, DataPtr, NULL);
    
    /* Allocate a BD from the transmit channel */
    taskLock();
    Status = TXBD_ALLOC(1, &BdPtr);
    taskUnlock();
    
    if (Status != XST_SUCCESS)
    {
        goto ReturnFatal;
    }

    /* Flush the cache for this mBlk */
    CACHE_DRV_FLUSH(&DriverPtr->cacheFuncs, DataPtr, PktLen);

    /* Setup BD */
    XDmaBdV3_mSetBufAddrLow(BdPtr, CACHE_DMA_VIRT_TO_PHYS(DataPtr));
    XDmaBdV3_mSetLength(BdPtr, PktLen);
    XDmaBdV3_mSetId(BdPtr, TX_COPY_BUF_ID);
    XDmaBdV3_mSetLast(BdPtr);

    /* Do CSUM offload setup */
#ifdef XTEA_CSUM_OFFLOAD
    if (DriverPtr->HwCaps.cap_enabled & IFCAP_TXCSUM)
    {
        CsumTx(MblkPtr, BdPtr);
    }
    else
    {
        XTemac_mSgSendBdCsumDisable(BdPtr);
    }
#endif
    
    /* Add the BD(s) to the SG engine */
    taskLock();
    Status = TXBD_TOHW(1, BdPtr);
    taskUnlock();

    if (Status == XST_SUCCESS)
    {
        COUNT_OUT_PKT(MblkPtr);
        netMblkClChainFree(MblkPtr);
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(OK);
    }

    ReturnFatal:
        XTemac_Stop(MAC);
        COUNT_OUT_ERROR(MblkPtr);
        netMblkClChainFree(MblkPtr);
        DriverPtr->VxEndLastError.errCode = END_ERR_DOWN;
        DriverPtr->VxEndLastError.pMesg = "Fatal SGDMA Tx error";
        muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 
        END_TX_SEM_GIVE(&DriverPtr->VxEnd);
        return(ERROR);
}


/******************************************************************************
*
* FUNCTION:
*
* SgSendInt
*
* DESCRIPTION:
*
* Callback function (called from driver) to handle confirmation of frame 
* transmit in SGDMA mode
*
* Processing of the buffer descriptors is deferred to task level.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* None.
*
******************************************************************************/
static void SgSendInt(XTemacEnd *DriverPtr)
{    
    /* Disable further send interrupts */
    XTemac_IntrSgDisable(MAC, XTE_SEND);

    DEBUG_LOG(DEBUG_TX, "SgSendInt: called\n", 1,2,3,4,5,6);

    /* Defer processing to task level */
    if (netJobAdd((FUNCPTR)SgSendTask, (int)DriverPtr, 0,0,0,0) != 0)
    {
        DEBUG_LOG(DEBUG_ERROR,"SgSendInt: netJobAdd failed, errno=%d\n",
                  errno,2,3,4,5,6);
    }
}

/******************************************************************************
*
* FUNCTION:
*
* SgSendTask
*
* DESCRIPTION:
*
* Task level handler for processing transmitted frames in SGDMA mode.
*
* This function is responsible for returning system resources after frame(s)
* have been sent.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void SgSendTask(XTemacEnd *DriverPtr)
{
    XDmaBdV3 *BdPtr;
    XDmaBdV3 *BdCurPtr;
    Xuint32 NumBds;
    Xuint32 TrackingId;
    XStatus Status;
    int i;

    /* Continue to process BDs while they are available */
    taskLock();
    NumBds = TXBD_FROMHW(0xFFFFFFFF, &BdPtr);
    taskUnlock();
    
    while (NumBds > 0)
    {
        DEBUG_LOG(DEBUG_TX, "SgSendTask: NumBds=%d\n",NumBds,2,3,4,5,6);

        /* Loop through each descriptor, retrieve the tracking ID and free
         * associated resources
         */
        for (i=0, BdCurPtr=BdPtr; i<NumBds; i++)
        {
            DEBUG_LOG(DEBUG_TX, "SgSendTask: Bd=%08X, buf=%08X, len=%d\n",
                      (int)BdCurPtr, XDmaBdV3_mGetBufAddrLow(BdCurPtr),
                      XDmaBdV3_mGetLength(BdCurPtr), 4,5,6);

            TrackingId = XDmaBdV3_mGetId(BdCurPtr);
            
            if (TrackingId == TX_COPY_BUF_ID)
            {
                TX_COPY_BUF_FREE(DriverPtr);            
            }
            else if (XDmaBdV3_mIsLast(BdCurPtr))
            {
                netMblkClChainFree((M_BLK_ID)TrackingId);
            }
            
            BdCurPtr = TXBD_NEXT(BdCurPtr);
        }

        /* Free BDs just processed */
        taskLock();
        Status = TXBD_FREE(NumBds, BdPtr);
        taskUnlock();
        
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR, "SgSendTask: Error freeing BDs, code %d\n",
                      Status,2,3,4,5,6);
            XTemac_Stop(MAC);
            DriverPtr->VxEndLastError.errCode = END_ERR_DOWN;
            DriverPtr->VxEndLastError.pMesg = "Fatal SGDMA Tx error";
            muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 
        }

        /* If a DMA interrupt had been pending, clear it */
        TXBD_ACK();

        /* Double check BD availability to avoid race */
        taskLock();
        NumBds = TXBD_FROMHW(0xFFFFFFFF, &BdPtr);
        taskUnlock();
    }

    /* If the send channel was blocked due to insufficient resources, unblock */
    if (DriverPtr->TxIsBlocked)
    {
        DEBUG_LOG(DEBUG_TX, "SgSendTask: Unblocked send\n",1,2,3,4,5,6);
        
        DriverPtr->TxIsBlocked = XFALSE;
        if (netJobAdd((FUNCPTR)muxTxRestart, (int)&DriverPtr->VxEnd, 0, 0,
                        0, 0) != OK)
        {
            DEBUG_LOG(DEBUG_ERROR,"SgSendTask: netJobAdd failed, errno=%d\n",
                      errno,2,3,4,5,6);
        }
    }    

    /* Re-enable transmit complete interrupts */
    XTemac_IntrSgEnable(MAC, XTE_SEND);
}


/******************************************************************************
*
* FUNCTION:
*
* SgRecvInt
*
* DESCRIPTION:
*
* Callback function (called from driver) to handle frames received in
* scatter-gather DMA mode. 
*
* Processing of the buffer descriptors is deferred to task level.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
* RETURN VALUE:
*
* None.
*
* NOTES:
*
******************************************************************************/
static void SgRecvInt(XTemacEnd *DriverPtr)
{
    /* Disable further receive interrupts */
    XTemac_IntrSgDisable(MAC, XTE_RECV);

    DEBUG_LOG(DEBUG_RX, "SgRecvInt: Called\n", 1,2,3,4,5,6);
    
    /* Call the task level handler to process incoming frames */
    if (netJobAdd((FUNCPTR)SgRecvTask, (int)DriverPtr, 0,0,0,0) != 0)
    {
        DEBUG_LOG(DEBUG_ERROR,"SgRecvTask: netJobAdd failed, errno=%d\n",
                  errno,2,3,4,5,6);
    }
}


/******************************************************************************
*
* FUNCTION:
*
* SgRecvTask
*
* DESCRIPTION:
*
* Task level handler for processing frames received in scatter-gather DMA mode.
* A list of descriptors which may describe one or more received frames is
* processed and passed up the stack. 
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* The ID of each buffer descriptor is an MBlk.
* Each MBlk has the following previously setup (see SgServiceRxRing()):
*   - MBlk.mBlkHdr.mData already has an assigned cluster containing received
*     data
*   - MBlk.mBlkHdr.mFlags already has its M_EXT flag set indicating a cluster is
*     assigned.
*   - MBlk.mBlkHdr.mNext is NULL
*
* Access to SGDMA descriptor ring is single threaded on the receive channel
* so there is no critical regions to protect ring counters required.
******************************************************************************/
static void SgRecvTask(XTemacEnd* DriverPtr)
{
    M_BLK_ID MblkPtr;
    XDmaBdV3 *BdPtr;
    XDmaBdV3 *BdCurPtr;
    int i;
    int BuffersLow = 0;
    Xuint32 NumBds;
    XStatus Status;

    /* Continue to process BDs while they are available */
    while((NumBds = RXBD_FROMHW(0xFFFFFFFF, &BdPtr)) > 0)
    {
        DEBUG_LOG(DEBUG_RX, "SgRecvTask: NumBds=%d\n",NumBds,2,3,4,5,6);
    
        /* The following check looks at whether VxWorks is holding onto all our
         * buffers. We know this is the case when there are no buffers enqueued
         * to receive SGDMA engine and there are no clusters available.
         */
        if ((DriverPtr->RxBuf.NetPool.clTbl[0]->clNumFree == 0) &&
            (MAC->RecvDma.BdRing.HwCnt == 0))
        {
            BuffersLow = 1;
            DEBUG_LOG(DEBUG_ERROR, "SgRecvTask: Buffers low\n",1,2,3,4,5,6);
        }
        
        /* Buffer descriptor processing:
         *
         * Loop through all the buffer descriptors extracting the associated MBlk
         * and linking them into packets. When a packet has been assembled, then
         * pass it up the stack.
         */
        for (i=1, BdCurPtr=BdPtr; i<=NumBds; i++)
        {
            /* Prepare the Mblk to pass up the stack */
            MblkPtr = (M_BLK_ID)XDmaBdV3_mGetId(BdCurPtr);
            
            /* If a low buffer condition exists, then do not pass the last buffer
             * up the stack. We need this buffer to keep the receiver running.
             */
            if (BuffersLow && (i == NumBds))
            {
                COUNT_IN_DISCARD(MblkPtr);
                
#if (_WRS_VXWORKS_MAJOR >= 6)
                netTupleFree(MblkPtr);
#else
                netMblkClChainFree(MblkPtr);
#endif
                
#if defined(END_ERR_NO_BUF)
                DriverPtr->VxEndLastError.errCode = END_ERR_NO_BUF;
                DriverPtr->VxEndLastError.pMesg = "Out of Rx buffers";
                muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 
#endif  
            }
            else
            {
                /* Get info from mblk */
                MblkPtr->mBlkHdr.mLen = XDmaBdV3_mGetLength(BdCurPtr);
                MblkPtr->mBlkPktHdr.len = MblkPtr->mBlkHdr.mLen;
                MblkPtr->mBlkHdr.mFlags |= M_PKTHDR;
                
                DEBUG_LOG(DEBUG_RX, "SgRecvTask: Bd=%08X, mBlk=%08X, mData=%08X, len=%d\n",
                          (int)BdCurPtr, (int)MblkPtr, (int)MblkPtr->mBlkHdr.mData,
                          (int)MblkPtr->mBlkHdr.mLen, 5,6);
                
#ifdef XTEA_CSUM_OFFLOAD
                if (DriverPtr->HwCaps.cap_enabled & IFCAP_RXCSUM)
                {
                    /* Grab csum data from HW, if it is 0 then HW did not
                     * calculate it */
                    MblkPtr->mBlkPktHdr.csum_data = XTemac_mSgRecvBdCsumGet(BdCurPtr);
                    if (MblkPtr->mBlkPktHdr.csum_data)
                    {
                        MblkPtr->mBlkPktHdr.csum_flags = CSUM_DATA_VALID;
                    }
                    else
                    {
                        MblkPtr->mBlkPktHdr.csum_flags = 0;
                    }
                }
#endif
                /* Pass up stack */
                COUNT_IN_PKT(MblkPtr);
                END_RCV_RTN_CALL(&DriverPtr->VxEnd, MblkPtr);
            }
            
            /* Get the next BD in the list */
            BdCurPtr = RXBD_NEXT(BdCurPtr);
        }
    
        /* Free BDs processed above */
        Status = RXBD_FREE(NumBds, BdPtr);
        
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_TX, "SgRecvTask: Error freeing BDs\n",1,2,3,4,5,6);
            COUNT_IN_ERROR();
            DriverPtr->VxEndLastError.errCode = END_ERR_WARN;
            DriverPtr->VxEndLastError.pMesg = "SGDMA Rx ring free error";
            muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 
        }

        /* If a DMA interrupt had been pending, clear it */
        RXBD_ACK();
    }

    /* Housekeeping */
    SgServiceRxRing(DriverPtr);

    /* Enable receive interrupts */
    XTemac_IntrSgEnable(MAC, XTE_RECV);
}


/******************************************************************************
*
* FUNCTION:
*
* SgServiceRxRing
*
* DESCRIPTION:
*
* Allocate and attach as many receive buffers to buffer descriptors as the
* descriptor ring will contain.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
******************************************************************************/
static void SgServiceRxRing(XTemacEnd *DriverPtr)
{
    XDmaBdV3 *BdPtr;
    XDmaBdV3 *BdCurPtr;
    XStatus Status;
    M_BLK_ID MblkPtr;
    int FreeBds, FreeBufs, CommitBds;
    int i;
    
    /* Get the number of free resources. Since this is the only place that
     * "consumes" receive channel BDs and buffers, get the current count of
     * each that are free to enqueue to HW.
     *
     * The number of BDs to prepare is the lesser of the number of BDs available
     * and the number of Mblks available.
     */
    FreeBds = RXBD_FREECNT();
    FreeBufs = DriverPtr->RxBuf.NetPool.clTbl[0]->clNumFree;
    CommitBds = min(FreeBds, FreeBufs);
                    
    if (CommitBds == 0)
    {
        return;
    }

    /* Allocate from the receive channel */
    Status = RXBD_ALLOC(CommitBds, &BdPtr);

    /* There should be enough BDs, getting an error here is not expected */
    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "SgServiceRxRing: Alloc error, code %d\n",
                  Status, 2,3,4,5,6);
        COUNT_IN_ERROR();
        return;
    }
    
    for(i=0, BdCurPtr=BdPtr; i<CommitBds; i++)
    {
        /* Get an entire Mblk tuple (Mblk, Cblk, and cluster) */
        MblkPtr = netTupleGet(&DriverPtr->RxBuf.NetPool,
                              DriverPtr->RxBuf.ClSize,
                              M_DONTWAIT, MT_DATA, FALSE);

        /* Out of buffers? This is not expected */
        if (MblkPtr == NULL)
        {
            SgRecoverRxMblks(DriverPtr, CommitBds-i, BdPtr);
            RXBD_UNALLOC(CommitBds, BdPtr);
            DEBUG_LOG(DEBUG_ERROR,"SgServiceRxRing: Null Mblk returned\n",
                      1,2,3,4,5,6);
            COUNT_IN_ERROR();
            return;
        }

        /* Invalidate cache for this buffer */
        CACHE_DRV_INVALIDATE(&DriverPtr->cacheFuncs, MblkPtr->mBlkHdr.mData,
                             DriverPtr->MaxFrameSz);

        /* Setup BD */
        XDmaBdV3_mSetLength(BdCurPtr, DriverPtr->MaxFrameSz);
        XDmaBdV3_mSetBufAddrLow(BdCurPtr,
                                CACHE_DMA_VIRT_TO_PHYS(MblkPtr->mBlkHdr.mData));
        XDmaBdV3_mSetId(BdCurPtr, MblkPtr);

        DEBUG_LOG(DEBUG_RX, "SgServiceRxRing: Bd=%08X, mBlk=%08X, mData=%08X\n",
                  (int)BdCurPtr, (int)MblkPtr, (int)MblkPtr->mBlkHdr.mData, 4,5,6);

        BdCurPtr = RXBD_NEXT(BdCurPtr);
    }

    /* Enqueue BDs to HW */
    Status = RXBD_TOHW(CommitBds, BdPtr);

    if (Status != XST_SUCCESS)
    {
        DEBUG_LOG(DEBUG_ERROR, "SgServiceRxRing: Commit error code %d\n",
                  Status, 2,3,4,5,6);
        COUNT_IN_ERROR();

        /* Recover resources */
        SgRecoverRxMblks(DriverPtr, CommitBds, BdPtr);
        RXBD_UNALLOC(CommitBds, BdCurPtr);
        
        /* Report this error */
        DriverPtr->VxEndLastError.errCode = END_ERR_WARN;
        DriverPtr->VxEndLastError.pMesg = "SGDMA Rx commit ring error";
        muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError);        
    }
}

/******************************************************************************
*
* FUNCTION:
*
* SgRecoverRxMblks
*
* DESCRIPTION:
*
* Extract and free Mblks from the RxBD list.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
* NumBds is the number of BDs to process
* BdPtr is the first BD to process
*
******************************************************************************/
static void SgRecoverRxMblks(XTemacEnd *DriverPtr, int NumBds, XDmaBdV3 *BdPtr)
{
    XDmaBdV3* BdCurPtr = BdPtr;
    int BdsLeft = NumBds;

    while(BdsLeft--)
    {
#if (_WRS_VXWORKS_MAJOR >= 6)
        netTupleFree((M_BLK_ID)XDmaBdV3_mGetId(BdCurPtr));
#else
        netMblkClChainFree((M_BLK_ID)XDmaBdV3_mGetId(BdCurPtr));
#endif
        BdCurPtr = RXBD_NEXT(BdCurPtr);
    }
}

/******************************************************************************
*
* FUNCTION:
*
* MemorySgDmaInit
*
* DESCRIPTION:
*
* Allocate and initialize the descriptor memory space.
*
* We currently assume that any memory space passed into the adapter for use
* with descriptors is non-cacheable.  Likewise, if memory is not provided to the
* adapter, the adapter allocates cache coherent memory for descriptors. It does
* this using the VxWorks cacheDmaMalloc routine. 
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
* InitDataPtr is a pointer to parsed initstring data
*
* RETURN VALUE:
*
*   XST_SUCCESS                 Space was allocated and initialized
*                               successfully
*   XST_EMAC_MEMORY_SIZE_ERROR  Space is not big enough to hold the
*                               minimum number of buffer descriptors
*   XST_EMAC_MEMORY_ALLOC_ERROR Problem with supplied space or allocation error
*   XST_DMA_ERROR               Error giving buffer descriptors to driver
*
* NOTES:
*
* Caching of descriptors is not currently supported.  There are two reasons
* for not supporting cached descriptors: (1) Multiple descriptors may fit into
* a single cache line. When a descriptor address needs to be flushed or
* invalidated, the entire cache line is flushed or invalidated, so nearby
* descriptors that may be in use by the hardware can be unintentially flushed
* or invalidated. (2) The Xilinx layered software framework, which facilitates
* portability of drivers, discourages the use of OS-dependent functions like
* cache invalidate/flush functions in lower layer software.  It is exactly this
* lower layer software, though, that manipulates buffer descriptors. The design
* and framework does not preclude the possibility of putting OS-dependent
* functionality in the lower layer software, but it is discouraged.
*
******************************************************************************/
static XStatus MemorySgDmaInit(XTemacEnd *DriverPtr, InitStringData *InitDataPtr)
{
    Xuint32 BdSpaceBytes = InitDataPtr->BdBytes;
    Xuint32 RxBdCnt = InitDataPtr->RxBdCnt;
    Xuint32 TxBdCnt = InitDataPtr->TxBdCnt;
    Xuint32 MemoryMinSize;
    Xuint32 RecvSize;
    Xuint32 SendSize;
    XStatus Status;
    XDmaBdV3 BdTemplate;
    
    /* Setup memory allocation data pointers in case of premature exit */
    DriverPtr->SgDma.IsBdMalloc = XFALSE;

    /* Save base address */
    DriverPtr->SgDma.BdSpaceAddr = InitDataPtr->BdBaseAddr;
    
    /* Set default descriptor count if not specified */
    if (RxBdCnt == 0)
    {
        RxBdCnt = XTEA_DFT_RXBDS;
    }
    if (TxBdCnt == 0)
    {
        TxBdCnt = XTEA_DFT_TXBDS;
    }
    
    /* Calculate the minimum memory requirements */
    MemoryMinSize = XDmaV3_mSgListMemCalc(DESC_ALIGNMENT, RxBdCnt + TxBdCnt);

    /* If no memory region size was given, then use the minimum. Otherwise make
     * sure that the size provided was big enough
     */
    BdSpaceBytes = InitDataPtr->BdBytes;
    if (BdSpaceBytes == 0)
    {
        BdSpaceBytes = MemoryMinSize;
    }
    else if (BdSpaceBytes < MemoryMinSize)
    {
        printf("temac: Descriptor space too small, must be at least 0x%08X bytes\n",
               (unsigned)MemoryMinSize);
        Status = XST_EMAC_MEMORY_SIZE_ERROR;
        goto ErrorExit;
    }

    /* If no memory region was specified in the initialization string, then
     * cache safe memory is allocated.
     *
     * If a memory region was specified in the initialization string, then
     * we assume that memory is cache safe. Alignment is checked.
     */
    if (DriverPtr->SgDma.BdSpaceAddr == 0)
    {
        /* cacheDmaMalloc should provide us with aligned space */
        DriverPtr->SgDma.BdSpaceAddr = (Xuint32)cacheDmaMalloc(BdSpaceBytes);

        if ((void*)DriverPtr->SgDma.BdSpaceAddr == NULL)
        {
            printf("temac: Error allocating memory for descriptors\n");
            Status = XST_EMAC_MEMORY_ALLOC_ERROR;
            goto ErrorExit;
        }
                                
        DriverPtr->SgDma.IsBdMalloc = XTRUE;
        
        /* Buffer descriptors must have cache coherent memory. The cacheability
         * buffer descriptor memory was set in the call to cacheDmaMalloc()
         * above.
         */
        if (!CACHE_DMA_IS_WRITE_COHERENT() || !CACHE_DMA_IS_READ_COHERENT())
        {
            printf("temac: Descriptor space requires cache coherent memory\n" );
            Status = XST_EMAC_MEMORY_ALLOC_ERROR;
            goto ErrorExit;
        }
    }
    else if (DriverPtr->SgDma.BdSpaceAddr & (DESC_ALIGNMENT - 1))
    {
        printf("temac: Descriptor space misaligned, must be on %d byte boundary\n",
               BUF_ALIGNMENT);
        Status = XST_EMAC_MEMORY_ALLOC_ERROR;
        goto ErrorExit;
    }

    /* Carve this block up into two parts, one for Rx descriptors and one for
     * Tx descriptors.
     */
    RecvSize = XDmaV3_mSgListMemCalc(DESC_ALIGNMENT, RxBdCnt);
    SendSize = XDmaV3_mSgListMemCalc(DESC_ALIGNMENT, TxBdCnt);

    /* Create the ring for Rx descriptors.
     * The following attributes will be in effect for all RxBDs
     */
    XDmaBdV3_mClear(&BdTemplate);
    XDmaBdV3_mSetLast(&BdTemplate);          /* 1:1 mapping of BDs to buffers */
    XDmaBdV3_mSetBufIncrement(&BdTemplate);  /* Buffers exist along incrementing
                                                addresses */
    XDmaBdV3_mIgnoreDre(&BdTemplate);        /* Don't need DRE for Rx channel */
    XDmaBdV3_mSetBdPage(&BdTemplate, 0);     /* Default to 32 bit addressing */
    XDmaBdV3_mSetBufAddrHigh(&BdTemplate, 0);/* Default to 32 bit addressing */
    XDmaBdV3_mSetDevSel(&BdTemplate, 0);     /* Always 0 */
    XDmaBdV3_mSetTransferType(&BdTemplate,   /* Data bus attributes */
                              XDMAV3_DMACR_TYPE_BFBURST_MASK,
                              XDMAV3_DMACR_DSIZE_64_MASK);
    
    Status = XTemac_SgSetSpace(MAC, XTE_RECV,
                               (Xuint32)CACHE_DMA_VIRT_TO_PHYS(DriverPtr->SgDma.BdSpaceAddr),
                               DriverPtr->SgDma.BdSpaceAddr,
                               DESC_ALIGNMENT,
                               RxBdCnt,
                               &BdTemplate);

    if (Status != XST_SUCCESS)
    {
        printf("temac: Error creating Rx BD space, code %d\n", (unsigned)Status);
        Status = XST_EMAC_MEMORY_ALLOC_ERROR;
        goto ErrorExit;
    }
    
    /* Create the ring for Tx descriptors. If no Tx DRE then buffers must occupy
     * a single descriptor, so set the "last" field for all descriptors.
     */
    XDmaBdV3_mClear(&BdTemplate);
    XDmaBdV3_mUseDre(&BdTemplate);           /* Always use DRE if available */
    XDmaBdV3_mSetBufIncrement(&BdTemplate);  /* Buffers exist along incrementing
                                                addresses */
    XDmaBdV3_mSetBdPage(&BdTemplate, 0);     /* Default to 32 bit addressing */
    XDmaBdV3_mSetBufAddrHigh(&BdTemplate, 0);/* Default to 32 bit addressing */
    XDmaBdV3_mSetDevSel(&BdTemplate, 0);     /* Always 0 */
    XDmaBdV3_mSetTransferType(&BdTemplate,   /* Data bus attributes */
                              XDMAV3_DMACR_TYPE_BFBURST_MASK,
                              XDMAV3_DMACR_DSIZE_64_MASK);
    XTemac_mSgSendBdCsumDisable(&BdTemplate); /* Disable csum offload by default */
    XTemac_mSgSendBdCsumSeed(&BdTemplate, 0); /* Don't need csum seed feature */
    
    if (!XTemac_mIsTxDre(MAC))
    {
        XDmaBdV3_mSetLast(&BdTemplate);
    }

    Status = XTemac_SgSetSpace(MAC, XTE_SEND,
                               (Xuint32)CACHE_DMA_VIRT_TO_PHYS(DriverPtr->SgDma.BdSpaceAddr + RecvSize),
                               DriverPtr->SgDma.BdSpaceAddr + RecvSize,
                               DESC_ALIGNMENT,
                               TxBdCnt,
                               &BdTemplate);

    if (Status != XST_SUCCESS)
    {
        printf("temac: Error creating Tx BD space, code %d\n", (unsigned)Status);
        Status = XST_EMAC_MEMORY_ALLOC_ERROR;
        goto ErrorExit;
    }
    
    /* Initialize the receive buffer descriptor ring in HW with new buffers */
    SgServiceRxRing(DriverPtr);

    /* This print format is a continuation of the data printed in MemoryInit */
#ifdef XTEA_PRINT_BUF_INIT_DATA
    printf("\t0x%08X bytes allocated for all BDs\n", (unsigned)BdSpaceBytes);
    printf("\t%5d RxBDs, BD space begins at 0x%08X\n",
           (unsigned)RxBdCnt,
           (unsigned)DriverPtr->SgDma.BdSpaceAddr);
    printf("\t%5d TxBDs, BD space begins at 0x%08X\n",
           (unsigned)TxBdCnt,
           (unsigned)(DriverPtr->SgDma.BdSpaceAddr + RecvSize));
#endif
    
    /* Success! */
    return(XST_SUCCESS);

    /* an error occurred somewhere, release any memory previously allocated */
ErrorExit:
    if (DriverPtr->SgDma.IsBdMalloc)
    {
        cacheDmaFree((void*)DriverPtr->SgDma.BdSpaceAddr);
    }
    
    return(Status);
}


/*****************************************************************************/
/*****************************************************************************
 * Utility, control and error handling functions common for all device types
 *****************************************************************************/
/*****************************************************************************/


/******************************************************************************
*
* FUNCTION:
*
* Mblk2Fifo
*
* DESCRIPTION:
*
* Copy the given Mblk chain into the packet FIFO. It is assumed the caller has
* checked for room in the FIFO and that the thread of execution has exclusive
* control over the transmit channel.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* MblkPtr is the VxWorks Mblk data chain to write into the packet FIFO
*
* RETURN VALUE:
*
*   XST_SUCCESS if the Mblk chain was successfully copied into the packet FIFO,
*   otherwise an error code from XTemac_FifoWrite or XTemac_FifoSend().
*
******************************************************************************/
static XStatus Mblk2Fifo(XTemacEnd *DriverPtr, M_BLK_ID MblkPtr)
{
    M_BLK_ID m = MblkPtr;
    XStatus Status;
    
    /* Loop through all Mblks for this packet */
    do
    {
        if (m->mBlkHdr.mNext != NULL)
        {
            /* More data in this Mblk chain, copy it to the FIFO */
            Status = XTemac_FifoWrite(MAC, m->mBlkHdr.mData, m->mBlkHdr.mLen,
                                      XTE_PARTIAL_PACKET);
            m = m->mBlkHdr.mNext;
        }
        else
        {
            /* Reached end of the packet, finish copy for this Mblk chain */
            Status = XTemac_FifoWrite(MAC, m->mBlkHdr.mData, m->mBlkHdr.mLen,
                                      XTE_END_OF_PACKET);

            /* Initiate transmit */
            if (Status == XST_SUCCESS)
            {
                Status = XTemac_FifoSend(MAC, MblkPtr->mBlkPktHdr.len);
            }
            
            break;
        }
    } while(Status == XST_SUCCESS);

    return(Status);
}


/******************************************************************************
*
* FUNCTION:
*
* Ioctl
*
* DESCRIPTION:
*
* Configure the device based on the possible END ioctl commands.  END ioctl
* commands are summarized below:
*
*   EIOCSFLAGS      Set the device flags (flags are defined in net/if.h). The
*                   device is stopped prior to setting the flags, and restarted
*                   when the flags are set (if it was previously started).
*   EIOCGFLAGS      Get the device flags
*   EIOCSADDR       Set MAC address.  The device is stopped prior to setting the
*                   flags, and restarted when the flags are set (if it was
*                   previously started).
*   EIOCGADDR       Get MAC address
*   EIOCGMIB2       Get the MIB-II statistics from the END structure
*   EIOCGNAME       Get the device name
*   EIOCGHDRLEN     Get the header length (of an Ethernet frame in this case)
*   EIOCGFBUF       Get min 1st buffer for chaining (NOT currently supported)
*   EIOCGMWIDTH     Get the device memory width
*   EIOCMULTIADD    Add a multicast address (NOT currently supported)
*   EIOCMULTIDEL    Delete a multicast address (NOT currently supported)
*   EIOCMULTIGET    Get a multicast address (NOT currently supported)
*   EIOCSRCVCB      Set the protocol stack's receive callback function (NOT
*                   currently supported)
*   EIOCPOLLSTART   Start polling
*   EIOCPOLLSTOP    Stop polling
*
* Custom ioctl commands:
*
*   XTEA_EIOSPKTTW Set packet threshold and waitbound timer for SG DMA
*   XTEA_EIOGPKTTW Get packet threshold and waitbound timer for SG DMA
*   XTEA_EIOSSGEND Set usage of SGEND interrupt
*   XTEA_EIOGSGEND Get SGEND interrupt usage
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* Cmd is the action requested
*
* Data is the input or output argument, depending on the action requested
*
* RETURN VALUE:
*
*   OK          The command was executed successfully
*   ERROR       An error occurred executing the command
*   EINVAL      The command is not supported
*
* NOTES:
*
* VxWorks offers additional ioctl commands in its NPT driver model. These
* commands are not supported by this adapter.  The commands are:
*
*   EIOCGNPT:           Support NPT
*   EIOCGMCASTLIST:     Get multicast address list
*   EIOCGBCASTADDR:     Get broadcast address
*   EIOCQUERY:
*   EIOCSMIB2233:       Set MIB-II per RFC2233
*   EIOCGMIB2233:       Get MIB-II per RFC2233
*
******************************************************************************/
static int Ioctl(XTemacEnd *DriverPtr, int Cmd, caddr_t Data)
{
    XTemacEnd_SgPktIoctlData *PktData;
    XStatus Status;
    int Rc = OK;
#ifdef XTEA_CSUM_OFFLOAD
    END_CAPABILITIES *HwCaps;
#endif
    
    switch ((unsigned)Cmd)
    {
        /* These commands require a change to the state of the device */
        case EIOCSFLAGS:
        case EIOCSADDR:
        case EIOCPOLLSTART:
        case EIOCPOLLSTOP:
        case XTEA_EIOCSPKTTW:
        case XTEA_EIOCSMPSPD:
            Rc =  IoctlSetDevControl(DriverPtr, Cmd, Data);
            break;

        /* Get the device options and convert them to VxWorks net/if.h
         * flags. Update the END structure just for safety.
         */
        case EIOCGFLAGS:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGFLAGS, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            *(long *)Data = END_FLAGS_GET(&DriverPtr->VxEnd);
            break;
            
        /* Get the MAC address */
        case EIOCGADDR:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGADDR, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            bcopy((char *)END_HADDR(&DriverPtr->VxEnd), (char *)Data,
                  END_HADDR_LEN(&DriverPtr->VxEnd));
            break;

        /* Get device name */            
        case EIOCGNAME:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGNAME, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            strcpy((char *)Data, XTE_DEVICE_NAME);
            break;

        /* Get the size of the Ethernet header. VxWorks uses this
         * command to know how many bytes to reserve at the start
         * of a buffer.
         */
        case EIOCGHDRLEN:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGHDRLEN, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            *(int *)Data = XTE_HDR_SIZE;
            break;


        /* Get the MIB-II counters.  The driver doesn't support MIB-II
         * currently, but VxWorks does use the MIB table for other device
         * information, such as the type of interface (Ethernet) and
         * MTU size
         */
        case EIOCGMIB2:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGMIB2, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            bcopy((char *)&DriverPtr->VxEnd.mib2Tbl, (char *)Data,
                  sizeof(DriverPtr->VxEnd.mib2Tbl));
            break;

        /* Get device memory width (not sure how this is used, it does not
         * appear to affect network buffer allocation).
         */
        case EIOCGMWIDTH:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGMWIDTH, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            *(int *)Data = NONE;   /* no access restrictions */
            break;

        /* Are we a NPT toolkit device */    
        case EIOCGNPT:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGNPT, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            Rc = EINVAL;
            break;

        /* Set receive callback */
        case EIOCSRCVCB:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCSRCVCB, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            Rc = EINVAL;
            break;

        /* Add multicast address */
        case EIOCMULTIADD:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCMULTIADD, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }
            Rc = MCastAddrAdd(&DriverPtr->VxEnd, (char*)Data);
            break;

        /* Get multicast address */
        case EIOCMULTIGET:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCMULTIGET, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }
            Rc = MCastAddrGet(&DriverPtr->VxEnd, (MULTI_TABLE*)Data);
            break;

        /* Delete multicast address */
        case EIOCMULTIDEL:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCMULTIDEL, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }
            Rc = MCastAddrDel(&DriverPtr->VxEnd, (char*)Data);
            break;

        /* Get minimum first buffer for chaining (?) */
        case EIOCGFBUF:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCGFBUF, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            Rc = EINVAL;
            break;

#ifdef XTEA_CSUM_OFFLOAD
        case EIOCGIFCAP:
            HwCaps = (END_CAPABILITIES *)Data;
            if (HwCaps == NULL)
            {
                Rc = EINVAL;
                break;
            }
            memcpy(HwCaps, &DriverPtr->HwCaps, sizeof(END_CAPABILITIES));
            break;
            
        case EIOCSIFCAP:
            HwCaps = (END_CAPABILITIES *)Data;
            if (HwCaps == NULL)
            {
                Rc = EINVAL;
                break;
            }
            DriverPtr->HwCaps.cap_enabled = HwCaps->cap_enabled;
            break;
#endif
            
        /* Get packet threshold and waitbound settings */    
        case XTEA_EIOCGPKTTW:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: XTEA_EIOGPKTTW, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (!XTemac_mIsSgDma(MAC) || (Data == NULL))
            {
                Rc = EINVAL;
                break;
            }

            PktData = (XTemacEnd_SgPktIoctlData*)Data;
            Status  = XTemac_IntrSgCoalGet(MAC, XTE_RECV,
                                           &PktData->RxPktThresholdCnt,
                                           &PktData->RxPktWaitbound);
            Status |= XTemac_IntrSgCoalGet(MAC, XTE_SEND,
                                           &PktData->TxPktThresholdCnt,
                                           &PktData->TxPktWaitbound);
            if (Status != XST_SUCCESS)
            {
                Rc = ERROR;
            }
            break;
            

        default:
            DEBUG_LOG(DEBUG_ERROR, "Ioctl: Bad request %08X, Data=%08X\n",
                      Cmd, (int)Data,3,4,5,6);
            return(EINVAL);
    }

#ifdef XMAC_DEBUG
    if (Rc != OK)
    {
        DEBUG_LOG(DEBUG_ERROR,"Ioctl: Processing error, code %d\n", Rc,2,3,4,5,6);
    }
#endif
    
    return(Rc);
}


/******************************************************************************
*
* FUNCTION:
*
* IoctlSetDevControl
*
* DESCRIPTION:
*
* Called by local function ioctl, this function updates devices settings that
* require the interface be stopped during the update.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
* Cmd is the action requested
* Data is the input or output argument, depending on the action requested
*
* RETURN VALUE:
*
*   OK          The command was executed successfully
*   ERROR       An error occurred executing the command
*   EINVAL      The command is not supported
*
* NOTES:
*
* See ioctl function for detailed information on device control commands.
*
******************************************************************************/
static int IoctlSetDevControl(XTemacEnd *DriverPtr, int Cmd, caddr_t Data)
{
    int Rc = OK;
    Xuint32 OptionsToSet, OptionsToClear;
    long Value, OldFlags, NewFlags;
    XTemacEnd_SgPktIoctlData *PktData;
    XStatus Status;
    Xboolean DeviceRunning = XTemac_mIsStarted(MAC);

    /* Get the configuration semaphore */
    semTake(DriverPtr->SemConfig, WAIT_FOREVER);

    /* Stop device if it is running */
    if (DeviceRunning)
    {
        XTemac_Stop(MAC);
    }

    OptionsToSet = 0;
    OptionsToClear = 0;
    
    /* Perform configuration operation */
    switch(Cmd)
    {
        /* Set device flags.  These flags are those defined by VxWorks,
         * so we need to translate them to the appropriate device options.
         */
        case EIOCSFLAGS:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCSFLAGS, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            
            OldFlags = END_FLAGS_GET(&DriverPtr->VxEnd);
            Value = (long)Data;

            /* Negative argument clears flags */
            if (Value < 0)
            {
                Value = -Value;
                Value--;
                END_FLAGS_CLR(&DriverPtr->VxEnd, Value);
            }
            else
            {
                END_FLAGS_SET(&DriverPtr->VxEnd, Value);
            }

            /* Handle flags set or cleared by the command */
            NewFlags = END_FLAGS_GET(&DriverPtr->VxEnd);

            /* IFF_BROADCAST */
            if ((NewFlags & IFF_BROADCAST) && !(OldFlags & IFF_BROADCAST))
            {
                OptionsToSet |= XTE_BROADCAST_OPTION;
            }
            else if ((OldFlags & IFF_BROADCAST) && !(NewFlags & IFF_BROADCAST))
            {
                OptionsToClear |= XTE_BROADCAST_OPTION;
            }
            
            /* IFF_PROMISC */
            if ((NewFlags & IFF_PROMISC) && !(OldFlags & IFF_PROMISC))
            {
                OptionsToSet |= XTE_PROMISC_OPTION;
            }
            else if ((OldFlags & IFF_PROMISC) && !(NewFlags & IFF_PROMISC))
            {
                OptionsToClear |= XTE_PROMISC_OPTION;
            }

            /* IFF_MULTICAST */
            if ((NewFlags & IFF_MULTICAST) && !(OldFlags & IFF_MULTICAST))
            {
                OptionsToSet |= XTE_MULTICAST_CAM_OPTION;
            }
            else if ((OldFlags & IFF_MULTICAST) && !(NewFlags & IFF_MULTICAST))
            {
                OptionsToClear |= XTE_MULTICAST_CAM_OPTION;
            }

            break;
            
        /* Set MAC address.  Store the input address in the MIB structure,
         * and give it to the driver.
         */
        case EIOCSADDR:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCSADDR, Data=%08X\n", (int)Data,
                      2,3,4,5,6);
            if (Data == NULL)
            {
                Rc = EINVAL;
                break;
            }

            /* Copy it to the MIB structure for VxWorks, then set it in the
             * device
             */
            bcopy((char *)Data, (char *)END_HADDR(&DriverPtr->VxEnd),
                  END_HADDR_LEN(&DriverPtr->VxEnd));
            XTemac_SetMacAddress(MAC, (char *)Data);
            break;

        /* Place the device into POLLED mode */
        case EIOCPOLLSTART:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCPOLLSTART\n",1,2,3,4,5,6);
            if (XTemac_mIsFifo(MAC))
            {
                OptionsToSet |= XTE_POLLED_OPTION;
            }
            else
            {
                Rc = EINVAL;
            }
            break;

        /* Take the device out of POLLED mode */
        case EIOCPOLLSTOP:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: EIOCPOLLSTOP\n",1,2,3,4,5,6);
            if (XTemac_mIsFifo(MAC))
            {
                OptionsToClear |= XTE_POLLED_OPTION;
            }
            else
            {
                Rc = EINVAL;
            }
            break;

        /* Set packet threshold and waitbound settings */
        case XTEA_EIOCSPKTTW:
            DEBUG_LOG(DEBUG_IOCTL, "Ioctl: XTEA_EIOSPKTTW, Data=%08X\n",
                      (int)Data,2,3,4,5,6);
            if (!XTemac_mIsSgDma(MAC) || (Data == NULL))
            {
                Rc = EINVAL;
                break;
            }

            PktData = (XTemacEnd_SgPktIoctlData*)Data;

            Status  = XTemac_IntrSgCoalSet(MAC, XTE_RECV,
                                           PktData->RxPktThresholdCnt,
                                           PktData->RxPktWaitbound);
            Status |= XTemac_IntrSgCoalSet(MAC, XTE_SEND,
                                           PktData->TxPktThresholdCnt,
                                           PktData->TxPktWaitbound);
            if (Status != XST_SUCCESS)
            {
                Rc = ERROR;
            }
            break;

        /* Set the MAC-PHY MII data clock */
        case XTEA_EIOCSMPSPD:
            Value = (long)Data;  /* prevents compiler warning */
            XTemac_SetOperatingSpeed(MAC, (Xuint16)Value);
            taskDelay(sysClkRateGet()/2);
            break;

        default:
            DEBUG_LOG(DEBUG_ERROR, "IoctlSetDevControl: Bad request %08X, Data=%08X\n",
                      Cmd, (int)Data,3,4,5,6);
            Rc = EINVAL;
            break;
    }

    /* If options have been changed, then set them in the device */
    if (OptionsToSet)
    {
        XTemac_SetOptions(MAC, OptionsToSet);
    }
    
    if (OptionsToClear)
    {
        XTemac_ClearOptions(MAC, OptionsToClear);
    }

    /* Restore device state */
    if (DeviceRunning)
    {
        XTemac_Start(MAC);
    }

    /* Return the configuration semaphore and exit*/
    semGive(DriverPtr->SemConfig);
    return(Rc);
}


/******************************************************************************
*
* FUNCTION:
*
* ErrorInt
*
* DESCRIPTION:
*
* Callback function (called from driver) to handle errors. Most work is queued
* to task level. If a fatal error has occurred, then the device is stopped.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
* ErrorClass is the type of error
* Word1 is the first error specific word
* Word1 is the second error specific word
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* This function is called by the driver within interrupt context
*
******************************************************************************/
static void ErrorInt(XTemacEnd *DriverPtr, XStatus ErrorClass, Xuint32 Word1,
                     Xuint32 Word2)
{
    DEBUG_LOG(DEBUG_ERROR, "ErrorInt: Class=%d, Word1=%08X, Word2=%08X\n",
              ErrorClass, Word1, Word2, 4,5,6);
    
    switch(ErrorClass)
    {
        /* These are serious errors, the device must be stopped immediately */
        case XST_FIFO_ERROR:
        case XST_PFIFO_DEADLOCK:
        case XST_DMA_ERROR:
        case XST_IPIF_ERROR:
            XTemac_Stop(MAC);
            DEBUG_LOG(DEBUG_ERROR,"ErrorInt: Device stopped\n", 1,2,3,4,5,6);
            break;

        /* Non-fatal errors */
        case XST_RECV_ERROR:
        default:
            break;
    }
            
    /* Queue error processing to task level */
    netJobAdd((FUNCPTR)ErrorTask, (int)DriverPtr, (int)ErrorClass, (int)Word1,
              (int)Word2, 0);
}


/******************************************************************************
*
* FUNCTION:
*
* ErrorTask
*
* DESCRIPTION:
*
* Task level error handler. Report error to MUX and restart the device if
* warranted.
*
* ARGUMENTS:
*
* DriverPtr is the END driver instance
* ErrorClass is the type of error
* Word1 is the first error specific word
* Word1 is the second error specific word
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void ErrorTask(XTemacEnd* DriverPtr, XStatus ErrorClass, Xuint32 Word1,
                      Xuint32 Word2)
{
    DEBUG_LOG(DEBUG_ERROR,"ErrorTask: class=%d, word1=%08X, word2=%08X\n",
              ErrorClass, Word1, Word2, 4,5,6);

    switch(ErrorClass)
    {
        /* These are serious errors, the device must be restarted */
        case XST_IPIF_ERROR:
            DriverPtr->VxEndLastError.errCode = END_ERR_RESET;
            DriverPtr->VxEndLastError.pMesg = "IPIF error";
            break;
            
        case XST_FIFO_ERROR:
            DriverPtr->VxEndLastError.errCode = END_ERR_RESET;
            DriverPtr->VxEndLastError.pMesg = "Len/status FIFO error";
            break;
            
        case XST_PFIFO_DEADLOCK:
            DriverPtr->VxEndLastError.errCode = END_ERR_RESET;
            if (Word1 == XTE_RECV)
            {
                DriverPtr->VxEndLastError.pMesg = "Rx pkt FIFO deadlock";
            }
            else
            {
                DriverPtr->VxEndLastError.pMesg = "Tx pkt FIFO deadlock";
            }
            break;
            
        case XST_DMA_ERROR:
            DriverPtr->VxEndLastError.errCode = END_ERR_RESET;
            if (Word1 == XTE_RECV)
            {
                if (Word2 & XDMAV3_DMASR_DBE_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx DMA bus error";
                }
                else if (Word2 & XDMAV3_DMASR_DBT_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx DMA bus timeout";
                }
                else
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx DMA error";
                }
            }
            else
            {
                if (Word2 & XDMAV3_DMASR_DBE_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Tx DMA bus error";
                }
                else if (Word2 & XDMAV3_DMASR_DBT_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Tx DMA bus timeout";
                }
                else
                {
                    DriverPtr->VxEndLastError.pMesg = "Tx DMA error";
                }
            }
            break;
            
        /* Non-fatal errors */
        case XST_RECV_ERROR:
            DriverPtr->VxEndLastError.errCode = END_ERR_WARN;
            if (Word1 & XTE_IPXR_RECV_REJECT_MASK)
            {
                if (Word1 & XTE_IPXR_RECV_PFIFO_ABORT_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx frame rejected - pkt FIFO overflow";
                }
                else if (Word1 & XTE_IPXR_RECV_LFIFO_ABORT_MASK)
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx frame rejected - length FIFO overflow";
                }
                else
                {
                    DriverPtr->VxEndLastError.pMesg = "Rx frame rejected";
                }
            }
            else
            {
                DriverPtr->VxEndLastError.pMesg = "Rx unknown error";
            }
            break;
                
        default:
            DriverPtr->VxEndLastError.errCode = END_ERR_WARN;
            DriverPtr->VxEndLastError.pMesg = "Unknown error";
            break;
    }
            
    /* Notify the mux */
    muxError(&DriverPtr->VxEnd, &DriverPtr->VxEndLastError); 

    if (DriverPtr->VxEndLastError.errCode == END_ERR_RESET)
    {
        Restart(DriverPtr);
    }
}


/******************************************************************************
*
* FUNCTION:
*
* Restart
*
* DESCRIPTION:
*
* Configure, reset, and restart the MAC based on current settings. The device
* encountered an error that requires a reset to recover. Before resetting, the
* current configuration must be determined. The device is then reset and re-
* configured based on its current configuration. 
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* RETURN VALUE:
*
* None.
*
******************************************************************************/
static void Restart(XTemacEnd *DriverPtr)
{
    Xuint32 Options;
    Xuint16 Speed;
    XTemacEnd_SgPktIoctlData IntrCoal;
    long VxFlags;
    int i;
    Xuint8 McAddr[XTE_MULTI_CAM_ENTRIES][XTE_MAC_ADDR_SIZE];
    
    DEBUG_LOG(DEBUG_ERROR, "Restart: Initiating device reset\n", 1,2,3,4,5,6);

    /* The config semaphore is used around this entire function since the
     * Get/Sets are technically read/modify/writes and the Start requires the
     * semaphore as well.
     */
    semTake(DriverPtr->SemConfig, WAIT_FOREVER);

    /* Get current options and settings */
    Options = XTemac_GetOptions(MAC);
    Speed = XTemac_GetOperatingSpeed(MAC);
    
    for (i=0; i<XTE_MULTI_CAM_ENTRIES; i++)
    {
        XTemac_MulticastGet(MAC, McAddr[i], i);
    }
    
    if (XTemac_mIsSgDma(MAC))
    {
        XTemac_IntrSgCoalGet(MAC, XTE_SEND,
                             &IntrCoal.TxPktThresholdCnt,
                             &IntrCoal.TxPktWaitbound);
        XTemac_IntrSgCoalGet(MAC, XTE_RECV,
                             &IntrCoal.RxPktThresholdCnt,
                             &IntrCoal.RxPktWaitbound);
    }

    /* Reset the device. After reset the device is in a default configuration
     * with the transmitter and receiver disabled.
     */
    XTemac_Reset(MAC, XTE_NORESET_HARD);
    taskDelay(sysClkRateGet());
    
    /* Restore options and settings */
    XTemac_SetOptions(MAC, Options);
    XTemac_ClearOptions(MAC, ~Options);
    XTemac_SetMacAddress(MAC, (Xuint8*)END_HADDR(&DriverPtr->VxEnd));
    XTemac_SetOperatingSpeed(MAC, Speed);

    for (i=0; i<XTE_MULTI_CAM_ENTRIES; i++)
    {
        XTemac_MulticastAdd(MAC, McAddr[i], i);
    }
    
    /* Restore interrupt coalescing and enable specific interrupts sources
     * accordingly. The device will not generate any interrupts until
     * XTemac_Start() gets called
     */
    if (XTemac_mIsSgDma(MAC))
    {
        XTemac_IntrSgCoalSet(MAC, XTE_SEND,
                             IntrCoal.TxPktThresholdCnt,
                             IntrCoal.TxPktWaitbound);
        XTemac_IntrSgCoalSet(MAC, XTE_RECV,
                             IntrCoal.RxPktThresholdCnt,
                             IntrCoal.RxPktWaitbound);
    }

    
    /* Update the END flags */
    VxFlags = IFF_UP | IFF_RUNNING | IFF_NOTRAILERS | IFF_SIMPLEX;

    if (Options & XTE_BROADCAST_OPTION)
    {
        VxFlags |= IFF_BROADCAST;
    }

    if (Options & XTE_PROMISC_OPTION)
    {
        VxFlags |= IFF_PROMISC;
    }

    if (Options & XTE_MULTICAST_CAM_OPTION)
    {
        VxFlags |= IFF_MULTICAST;
    }
    
    END_OBJ_READY(&DriverPtr->VxEnd, VxFlags);
    
    /* For SGDMA, kick the transmit & receive channels to extract system
     * resources tied up in BDs that were ready to be post processed prior to
     * this reset event. This will also enable interrupts for both respective
     * channels.
     *
     * For FIFO direct, just enable interrupts.
     */
    if (XTemac_mIsSgDma(MAC))
    {
        netJobAdd((FUNCPTR)SgRecvTask, (int)DriverPtr, 0,0,0,0);
        netJobAdd((FUNCPTR)SgSendTask, (int)DriverPtr, 0,0,0,0);
    }
    else
    {
        XTemac_IntrFifoEnable(MAC, XTE_SEND | XTE_RECV);
    }

    /* Restart the device */
    XTemac_Start(MAC);
    semGive(DriverPtr->SemConfig);
}


/******************************************************************************
*
* FUNCTION:
*
* MemoryInit
*
* DESCRIPTION:
*
* Allocate and initialize the network buffer memory space. Receive buffers are
* allocated here and typically "loaned" (controlled by the BSP) to the network
* stack to avoid driver-to-stack buffer copying.  Transmit buffers are also
* allocated here for the sole purpose of stack-to-driver buffer copying.  The
* need for copying outgoing buffers depends on alignment restrictions in the
* device. See the comment block at the top of this file for more information.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
* InitDataPtr is a pointer to parsed initstring data
*
* RETURN VALUE:
*
*   XST_SUCCESS                     Space was allocated and initialized
*                                   successfully
*   XST_EMAC_MEMORY_ALLOC_ERROR     Space could not be allocated
*   XST_EMAC_MEMORY_SIZE_ERROR      Space is not big enough to hold the
*                                   minimum number of buffers
*
* NOTES:
*
* None.
*
******************************************************************************/
static XStatus MemoryInit(XTemacEnd *DriverPtr, InitStringData *InitDataPtr)
{
    Xuint32 BufSpaceBytes;
    Xuint32 MinAlign;
    Xuint32 MemoryMinSz;
    Xuint32 MemoryDftSz;
    Xuint32 GrossFrameSz;
    Xuint32 AdjustedBufSpaceAddr;
    unsigned ClFrameOverhead = sizeof(CL_BUF_ID);
    unsigned ClFixedOverhead = sizeof(int);
    CL_DESC ClDesc;
    XStatus Status;
    
    /* Setup memory allocation data pointers in case of premature exit */
    DriverPtr->RxBuf.ClConfig.memArea = NULL;
    DriverPtr->IsBufMalloc = XFALSE;

    /* Save base address */
    DriverPtr->BufSpaceAddr = InitDataPtr->BufBaseAddr;
    
    /* Calculate the required alignment for each frame buffer. HW has a minimum
     * requirement of BUF_ALIGNMENT, but typically we want to be aligned
     * along cache lines.
     */
    if (InitDataPtr->CacheLineSize > BUF_ALIGNMENT)
    {
        MinAlign = InitDataPtr->CacheLineSize;
    }
    else
    {
        MinAlign = BUF_ALIGNMENT;
    }

    /* Receive buffers will be managed by a VxWorks netpool. The netpool adds
     * overhead for each frame buffer as well as some amount of fixed overhead
     * (which I didn't find any documentation for but ln97xEnd.c uses it).
     *
     * GrossFrameSz is the number of bytes to contain a frame plus another
     * alignment term to contain any netpool overhead in its own cacheline.
     */
    GrossFrameSz = ROUND_UP(DriverPtr->MaxFrameSz + MinAlign, MinAlign);
    DriverPtr->RxBuf.ClSize = GrossFrameSz - ClFrameOverhead;

    /* Calculate the minimum and default memory sizes */
    MemoryMinSz = (XTEA_MIN_BUFS * GrossFrameSz) + ClFixedOverhead;
    MemoryDftSz = (XTEA_DFT_BUFS * GrossFrameSz) + ClFixedOverhead;

    /* If no memory region size was given, then use the default. Add in MinAlign
     * bytes in case we need to adjust alignment later. This ensures that we
     * always get at least MemoryDftSz bytes allocated.
     */
    if (InitDataPtr->BufBytes == 0)
    {
        BufSpaceBytes = MemoryDftSz + MinAlign;
    }
    else
    {
        BufSpaceBytes = InitDataPtr->BufBytes;
    }

    /* If no memory region was given, then cache safe memory is allocated.
     * If a memory region was given, then we assume it is cache coherent.
     */
    if (DriverPtr->BufSpaceAddr == 0)
    {
        /* Allocate cache safe memory */
        DriverPtr->BufSpaceAddr = (Xuint32)cacheDmaMalloc(BufSpaceBytes);

        if ((void*)DriverPtr->BufSpaceAddr == NULL)
        {
            printf("temac: Error allocating memory for buffers\n");
            Status = XST_EMAC_MEMORY_ALLOC_ERROR;
            goto ErrorExit;
        }
        DriverPtr->IsBufMalloc = XTRUE;
        DriverPtr->cacheFuncs = cacheDmaFuncs;
    }
    else
    {
        DriverPtr->cacheFuncs = cacheNullFuncs;
    }
                     
    /* Adjust BufSpaceAddr so that BufSpaceAddr + ClFrameOverhead lands on
     * MinAlign alignment
     */
    AdjustedBufSpaceAddr = ROUND_UP(DriverPtr->BufSpaceAddr + ClFrameOverhead,
                                    MinAlign) - ClFrameOverhead;

    /* Alignment process may have removed usable bytes from the total space */
    BufSpaceBytes -= (AdjustedBufSpaceAddr - DriverPtr->BufSpaceAddr);
    
    /* Verify the space is big enough for the minimum number of buffers */
    if (BufSpaceBytes < MemoryMinSz)
    {
        printf("temac: Specified buffer space is too small\n");
        Status = XST_EMAC_MEMORY_SIZE_ERROR;
        goto ErrorExit;
    }

    /* Calculate how many buffers we can fit into this space */
    DriverPtr->RxBuf.BufCnt = (BufSpaceBytes - ClFixedOverhead) / GrossFrameSz;

    /* Do we need to reserve buffers for stack-to-driver copying on the Tx
     * channel? Yes if SGDMA. There are two scenarios:
     *
     *   1. With Tx DRE, we will need to defrag any mblk chains greater than
     *      XTEA_MAX_DEFRAG_BUFS. The number of buffers to reserve for this
     *      is set to XTEA_MAX_DEFRAG_BUFS. This is not due to a HW limitation,
     *      just so we don't use up all our BDs on a handful of frames.
     *
     *   2. Without Tx DRE, every mblk chain must be copied so we need more
     *      buffers typically than case one. The number is controlled by the
     *      constant XTEA_COPY_BUFS_RATIO.
     */
    if (XTemac_mIsSgDma(MAC))
    {
        if (XTemac_mIsTxDre(MAC))
        {
            DriverPtr->TxBuf.BufCnt = XTEA_MAX_DEFRAG_BUFS;
            DriverPtr->RxBuf.BufCnt -= XTEA_MAX_DEFRAG_BUFS;

            if (DriverPtr->RxBuf.BufCnt <= 0)
            {
                printf("temac: Error allocated 0 Rx buffers\n");
                Status = XST_EMAC_MEMORY_SIZE_ERROR;
                goto ErrorExit;
            }
        }
        else
        {
            /* Redistribute buffers according to ratio */
            DriverPtr->TxBuf.BufCnt = DriverPtr->RxBuf.BufCnt /
                (XTEA_COPY_BUFS_RATIO + 1);
            DriverPtr->RxBuf.BufCnt -= DriverPtr->TxBuf.BufCnt;
            
            /* Check for errors */
            if (DriverPtr->TxBuf.BufCnt == 0)
            {
                printf("temac: Error - allocated 0 Tx buffers\n");
                Status = XST_EMAC_MEMORY_SIZE_ERROR;
                goto ErrorExit;
            }
            if (DriverPtr->RxBuf.BufCnt == 0)
            {
                printf("temac: Error - allocated 0 Rx buffers\n");
                Status = XST_EMAC_MEMORY_SIZE_ERROR;
                goto ErrorExit;
            }
        }
    }

    /* The number of Rx buffers in now known. Setup the netpool. This process 
     * is documented by Wind River.
     */
    ClDesc.clNum   = DriverPtr->RxBuf.BufCnt;
    ClDesc.clSize  = DriverPtr->RxBuf.ClSize;
    ClDesc.memArea = (char *)AdjustedBufSpaceAddr;    
    ClDesc.memSize = (ClDesc.clNum * GrossFrameSz) + ClFixedOverhead;

    DriverPtr->RxBuf.ClConfig.clBlkNum = ClDesc.clNum;
    DriverPtr->RxBuf.ClConfig.mBlkNum  = ClDesc.clNum * 2;

    DriverPtr->RxBuf.ClConfig.memSize =
        (DriverPtr->RxBuf.ClConfig.mBlkNum * (M_BLK_SZ + sizeof(long))) +
        (DriverPtr->RxBuf.ClConfig.clBlkNum * (CL_BLK_SZ + sizeof(long)));

    DriverPtr->RxBuf.ClConfig.memArea =
        memalign(sizeof(long), DriverPtr->RxBuf.ClConfig.memSize);
    
    if (DriverPtr->RxBuf.ClConfig.memArea == NULL)
    {
        printf("temac: Error allocating mBlk memory area\n");
        Status = XST_EMAC_MEMORY_ALLOC_ERROR;
        goto ErrorExit;
    }

    /* Hand over our memory to the netpool */
    if (netPoolInit(&DriverPtr->RxBuf.NetPool, &DriverPtr->RxBuf.ClConfig,
                    &ClDesc, 1, NULL) == ERROR)
    {
        printf("temac: Error allocating network pool\n");
        Status = XST_EMAC_MEMORY_ALLOC_ERROR;
        goto ErrorExit;
    }

    DriverPtr->VxEnd.pNetPool = &DriverPtr->RxBuf.NetPool;
    
#ifdef XTEA_PRINT_BUF_INIT_DATA
    printf("temac: Buffer information:\n");
    printf("\t0x%08X bytes allocated for all buffers\n", (unsigned)BufSpaceBytes);
    printf("\t%5d byte cluster size\n",
           (unsigned)(GrossFrameSz-ClFrameOverhead));
    printf("\t%5d Rx buffers, 1st buffer located at 0x%08X\n",
           DriverPtr->RxBuf.BufCnt, (unsigned)(ClDesc.memArea + ClFrameOverhead));
#endif

    /* Setup the locally managed Tx copy buffer management structure.
     * These buffers will be placed after the receive buffer pool
     */
    if (DriverPtr->TxBuf.BufCnt)
    {
        DriverPtr->TxBuf.BaseAddr =
            (Xuint8*) ROUND_UP(ClDesc.memArea + ClDesc.memSize, MinAlign);
        
        DriverPtr->TxBuf.HighAddr = DriverPtr->TxBuf.BaseAddr +
            (GrossFrameSz * DriverPtr->TxBuf.BufCnt) - 1;
        
        DriverPtr->TxBuf.NextBuf = DriverPtr->TxBuf.BaseAddr;
        DriverPtr->TxBuf.BufSize = GrossFrameSz;
        DriverPtr->TxBuf.BufsAvailable = DriverPtr->TxBuf.BufCnt;

#ifdef XTEA_PRINT_BUF_INIT_DATA
        printf("\t%5d Tx buffers, 1st buffer located at 0x%08X\n",
               DriverPtr->TxBuf.BufCnt, (unsigned)DriverPtr->TxBuf.BaseAddr);
#endif
    }
           
    /* If this instance has a SG DMA engine, then more memory initialization
     * must be performed
     */
    if (XTemac_mIsSgDma(MAC))
    {
        Status = MemorySgDmaInit(DriverPtr, InitDataPtr);
        if (Status != XST_SUCCESS)
        {
            goto ErrorExit;
        }
    }

    /* success! */
    return(XST_SUCCESS);

    /* an error occurred somewhere, release any memory previously allocated */
ErrorExit:
    if (DriverPtr->IsBufMalloc)
    {
        cacheDmaFree((void*)DriverPtr->BufSpaceAddr);
    }
    
    if (DriverPtr->RxBuf.ClConfig.memArea != NULL)
    {
        free(DriverPtr->RxBuf.ClConfig.memArea);
    }

    return(Status);
}

    
/******************************************************************************
*
* FUNCTION:
*
* ProcessInitString
*
* DESCRIPTION:
*
* Parse the driver initialization string into driver info structure elements.
* See the comments at the top of this file for details on the string format.
*
* ARGUMENTS:
*
* DPtr is a pointer to the converted data fields parsed out of the InitString.
*
* InitString is the initialization string.
*
* RETURN VALUE:
*
* XST_SUCCESS when successful, or XST_EMAC_PARSE_ERROR if there is a
* problem parsing the string.
*
* NOTES:
*
* None.
*
******************************************************************************/
static XStatus ProcessInitString(InitStringData *Dptr, char *InitString)
{
    char *ErrMsgDft = "temac: InitString not formatted correctly\n";
    char *Token;
    char *Tmp;

    /* First parameter is the unit number (from muxLib.o) */
    Token = strtok_r(InitString, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->Unit = (int)strtoul(Token, NULL, 0);

    /* Second parameter is the device ID of the MAC being initialized */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->DeviceId = (Xuint16)strtoul(Token, NULL, 0);

    /* Third parameter is the address of the 6-byte MAC address */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->MacAddrPtr = (Xuint8 *)strtoul(Token, NULL, 0);

    /* verify the MAC address pointer is not null */
    if (Dptr->MacAddrPtr == (Xuint8 *)0)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }

    /* Fourth parameter is the network-buffer memory address */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->BufBaseAddr = strtoul(Token, NULL, 0);
        
    /* Fifth parameter is the network-buffer memory size */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->BufBytes = (Xuint32)strtoul(Token, NULL, 0);

    /* Verify that if a base address was provided, then a size was also provided */
    if ((Dptr->BufBaseAddr != 0) && (Dptr->BufBytes == 0))
    {
        printf("temac: Initstring <BufBytes> of 0 is illegal when <BufBytes> is non-0\n");
        return(XST_EMAC_PARSE_ERROR);
    }
    
    /* Sixth parameter is the descriptor memory space */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->BdBaseAddr = strtoul(Token, NULL, 0);
    
    /* Seventh parameter is the descriptor memory size */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->BdBytes = (Xuint32)strtoul(Token, NULL, 0);

    /* Verify that if a descriptor base address was provided, then a size was
     * also provided
     */
    if ((Dptr->BdBaseAddr != 0) && (Dptr->BdBytes == 0))
    {
        printf("temac: Initstring <BdBytes> of 0 is illegal when <BdBaseAddr> is non-0\n");
        return(XST_EMAC_PARSE_ERROR);
    }

    /* Eigth parameters is RxBdCnt */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->RxBdCnt = (Xuint32)strtoul(Token, NULL, 0);

    /* Ninth parameters is RxBdCnt */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->TxBdCnt = (Xuint32)strtoul(Token, NULL, 0);
    
    /* Tenth parameter is the processor cache line size */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->CacheLineSize = (Xuint32)strtoul(Token, NULL, 0);

    /* Eleventh parameter is the interrupt ID */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        printf(ErrMsgDft);
        return(XST_EMAC_PARSE_ERROR);
    }
    Dptr->IntrId = (Xuint32)strtoul(Token, NULL, 0);

    /* Twelfth parameter is the MTU */
    Token = strtok_r(NULL, XTEA_INIT_STRING_DELIMITER, &Tmp);
    if (Token == NULL)
    {
        /* Wasn't there, use default */
        Dptr->MTU = XTE_MTU;
    }
    else
    {
        /* Convert this field and range check */
        Dptr->MTU = (Xuint32)strtoul(Token, NULL, 0);

        /* Too big? */
        if (Dptr->MTU > (XTE_JUMBO_MTU + 18))
        {
            printf("temac: InitString <MTU> field of %d is too large\n",
                   (unsigned)Dptr->MTU);
            return(XST_EMAC_PARSE_ERROR);
        }

        /* Use default */
        else if (Dptr->MTU == 0)
        {
            Dptr->MTU = XTE_MTU;
        }
    }
    return(XST_SUCCESS);
}


/******************************************************************************
*
* FUNCTION:
*
* MCastAddrAdd
*
* DESCRIPTION:
*
* Add a multicast address to the multicast address list in the END_OBJ
* structure. The MUX manages the multicast list, but the add/delete/get
* functions are in the adapter in case the hardware or driver provides multicast
* filtering.
*
* ARGUMENTS:
*
* EndPtr is a pointer to the device-specific END_OBJ structure.
*
* AddressPtr is a pointer to the multicast address to be added to the list.
*
* RETURN VALUE:
*
* OK          The command was executed successfully
* ERROR       An error occurred executing the command
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS MCastAddrAdd(END_OBJ *EndPtr, char *AddressPtr)
{
    XTemacEnd *DriverPtr = (XTemacEnd*)EndPtr;
    char NullAddr[XTE_MAC_ADDR_SIZE];
    XStatus Status;
    int Rc, i;

    /* There is some question to the meaning of return values of etherMultiAdd()
     * We assume the following: ENETRESET if the address was added to the list,
     * OK if the address was already in the list, ERROR if an error occurred
     * adding to the list.
     */
    Rc = etherMultiAdd(&EndPtr->multiList, AddressPtr);

    /* Create 0:0:0:0:0:0 address. If we find one that matches then there is
     * a free slot in the cam 
     */
    for (i=0; i < XTE_MAC_ADDR_SIZE; i++) NullAddr[i] = 0;
    
    /* If this is a new address, then we need to write it to the HW CAM */
    if (Rc == ENETRESET)
    {
        /* See if the address is already present */
        if (MCastAddrFind(DriverPtr, AddressPtr) != -1)
        {
            return(OK);
        }
        
        /* Find a slot containing a NULL address, this is the insertion point */
        i = MCastAddrFind(DriverPtr, NullAddr);
        if (i == -1)
        {
            DEBUG_LOG(DEBUG_ERROR, "MCastAddrAdd: Cam is full\n", 1,2,3,4,5,6);
            etherMultiDel(&EndPtr->multiList, AddressPtr);
            return(ERROR);
        }

        /* Add the address */
        XTemac_Stop(MAC);
        Status = XTemac_MulticastAdd(MAC, AddressPtr, i);
        XTemac_Start(MAC);

        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR, "mCastAddrAdd: CAM access failed, code %d\n",
                      Status,2,3,4,5,6);
            etherMultiDel(&EndPtr->multiList, AddressPtr);
            return(ERROR);
        }
    }
    else if (Rc != OK)
    {
        return(ERROR);
    }

    return(OK);
}


/******************************************************************************
*
* FUNCTION:
*
* MCastAddrDel
*
* DESCRIPTION:
*
* Delete a multicast address from the multicast address list in the END_OBJ
* structure. Additionally, make sure this address is removed from the HW CAM.
*
* ARGUMENTS:
*
* EndPtr is a pointer to the device-specific END_OBJ structure.
*
* AddressPtr is a pointer to the multicast address to be deleted from the list.
*
* RETURN VALUE:
*
* OK          The command was executed successfully
* ERROR       An error occurred executing the command
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS MCastAddrDel(END_OBJ *EndPtr, char *AddressPtr)
{
    XTemacEnd *DriverPtr = (XTemacEnd*)EndPtr;
    XStatus Status;
    int Rc, i;

    /* There is some question to the meaning of return values of etherMultiDel()
     * We assume the following: ENETRESET if the address was found and deleted
     * from the list, OK if the address wasn't in the list, ERROR if an error
     * occurred deleting it from the list.
     */
    Rc = etherMultiDel(&EndPtr->multiList, AddressPtr);
    
    /* If address was deleted, then go about deleting it from the CAM as well */
    if (Rc == ENETRESET)
    {
        /* Search for the address */
        i = MCastAddrFind(DriverPtr, AddressPtr);
        if (i == -1)
        {
            DEBUG_LOG(DEBUG_ERROR, "MCastAddrDel: Cam address not found\n",
                      1,2,3,4,5,6);
            return(ERROR);
        }

        /* Remove from HW CAM */
        XTemac_Stop(MAC);
        Status = XTemac_MulticastClear(MAC, i);
        XTemac_Start(MAC);
        
        if (Status != XST_SUCCESS)
        {
            DEBUG_LOG(DEBUG_ERROR, "MCastAddrDel: CAM access failed, code %d\n",
                      Status,2,3,4,5,6);
            return(ERROR);
        }
    }
    else if (Rc != OK)
    {
        return(ERROR);
    }

    return(OK);
}

/******************************************************************************
*
* FUNCTION:
*
* MCastAddrGet
*
* DESCRIPTION:
*
* Get the multicast address list from the END_OBJ structure. The MUX manages
* the multicast list, but the add/delete/get functions are in the adapter in
* case the hardware or driver provides multicast filtering.  In our case we do
* not yet support filtering in the hardware.
*
* ARGUMENTS:
*
* EndPtr is a pointer to the device-specific END_OBJ structure.
*
* MultiTablePtr is a pointer to the multicast address list to be returned.
*
* RETURN VALUE:
*
* OK          The command was executed successfully
* ERROR       An error occurred executing the command
*
* NOTES:
*
* None.
*
******************************************************************************/
static STATUS MCastAddrGet(END_OBJ *EndPtr, MULTI_TABLE *MultiTablePtr)
{
    return(etherMultiGet(&EndPtr->multiList, MultiTablePtr));
}


/******************************************************************************
*
* FUNCTION:
*
* MCastAddrFind
*
* DESCRIPTION:
*
* Search the device CAM for the given address. Return the CAM index that
* contains the matching entry.
*
* ARGUMENTS:
*
* DriverPtr is a pointer to the device-specific END_OBJ structure.
*
* AddressPtr is the address to search for.
*
* RETURN VALUE:
*
* -1 if the given address does not match anything in the CAM, or
* [0..XTE_MULTI_CAM_ENTRIES-1] of the matching address
*
******************************************************************************/
static int MCastAddrFind(XTemacEnd *DriverPtr, char *AddressPtr)
{
    char CamData[XTE_MAC_ADDR_SIZE];
    int i;

    for (i=0; i < XTE_MULTI_CAM_ENTRIES; i++)
    {
        XTemac_MulticastGet(MAC, CamData, i);
        if (memcmp(AddressPtr, CamData, XTE_MAC_ADDR_SIZE) == 0)
        {
            return(i);
        }
    }
    
    return(-1);  /* Not found */
}


#ifdef XTEA_CSUM_OFFLOAD

/******************************************************************************
*
* FUNCTION:
*
* CsumTx
*
* DESCRIPTION:
*
* Setup the provided buffer descriptor with checksum offload data if the
* Mblk requests it.
*
* HW can calculate the checksum for the UDP or TCP portion of the IP packet.
* The descriptor is setup to tell HW that the calculation starts on the 1st
* byte of the IP payload with the insertion point dependent on whether this
* is a UDP or TCP packet.
* 
* It is assumed the stack has calculated the IP header checksum and inserted
* it into the correct position and inserted the IP pseudo-header checksum
* into the TCP or UDP checksum word.
*
* ARGUMENTS:
*
* MblkPtr is the 1st Mblk in the chain for a packet
*
* BdPtr is the 1st BD in the chain for a packet
*
******************************************************************************/
static void CsumTx(M_BLK_ID MblkPtr, XDmaBdV3 *BdPtr)
{
    Xuint16 StartOffset, InsertOffset;

    /* Checksum offload is enabled in the stack so see if this is a UDP or
     * TCP request
     */
    if (MblkPtr->mBlkPktHdr.csum_flags & CSUM_TCP)
    {
        InsertOffset = 0x10;
    }
    else if (MblkPtr->mBlkPktHdr.csum_flags & CSUM_UDP)
    {
        InsertOffset = 0x6;
    }
    else
    {
        XTemac_mSgSendBdCsumDisable(BdPtr);
        return;
    }

    /* Finalize starting and insertion byte offset and tell HW. No seed
     * is required. All descriptors seed attributes should have been set to
     * 0x0000 at END load time
     */
    StartOffset = SIZEOF_ETHERHEADER + CSUM_IP_HDRLEN(MblkPtr);
    InsertOffset += StartOffset;
    XTemac_mSgSendBdCsumEnable(BdPtr);
    XTemac_mSgSendBdCsumSetup(BdPtr, StartOffset, InsertOffset);
}

#endif  /* XTEA_CSUM_OFFLOAD */
