/******************************************************************************
*
*       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*       SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*       FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*       FOR A PARTICULAR PURPOSE.
*
*       (c) Copyright 2005-2006 Xilinx Inc.
*       All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xtemac_end_adapter.h
*
* The Xilinx Ethernet Tri-Mode 10/100/1000 Adapter component. This component
* interfaces between the XTemac component and VxWorks.  The adapter plugs into
* the VxWorks MUX interface.  It follows the Enhanced Network Driver (END) 
* model.
*
* This driver supports the native VxWorks stack present in versions 5.5 and
* 6.x.
*
* There are two externally visible functions, XTemacEndAdapter_Load() and
* XTemacEndAdapter_GetDriver(). The Load function is called from the Board
* Support Package (BSP) to initialize a network driver.  It is plugged into a
* driver table in configNet.h (endDrvTbl).  Each XEthernet MAC should have a
* corresponding driver entry in this table. The GetDriver function returns the
* underlying XTemac Level 1 driver object.
*
* The XTemacEndAdapter_Load() function sets up a VxWorks function table, part
* of the END_OBJ structure, which VxWorks uses to access the adapter and
* ultimately the driver.
*
* @note
*
* This adapter is intended to be VxWorks-dependent and processor-independent.
* It works with virtual and/or physical addresses, and handles cache coherency
* as necessary.  This component does not adhere to the Network Protocol
* Toolkit (NPT) driver model.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1.00a rmm  01/11/05 First release
* 2.00a rmm  11/21/05 Include HW fixes for MII and multicast access,
*                     added checksum offloading for VxWorks 6.x, removed half
*                     duplex support, removed XTEA_EIOCSSGEND and
*                     XTEA_EIOCGSGEND ioctl commands
* </pre>
******************************************************************************/

#ifndef XTEMAC_END_ADAPTER_H /* prevent circular inclusions */
#define XTEMAC_END_ADAPTER_H /* by using protection macros */

/***************************** Include Files *********************************/

#include <vxWorks.h>
#include <end.h>
#include <endLib.h>
#include <netBufLib.h>
#include <cacheLib.h>
#include "xbasic_types.h"
#include "xtemac.h"

/************************** Constant Definitions *****************************/

#define XTEA_INIT_STRING_DELIMITER   ":"

/** @name Defaults
 */
/*@{*/

/** Default number of buffers to allocate if not specified within the
 *  InitString. 
 */
#define XTEA_DFT_BUFS                64

/** Absolute minimum number of buffers to allow. Using this number of
 *  buffers will affect performance throughput. 
 */
#define XTEA_MIN_BUFS                16

/** Default number of Rx buffer descriptors if SGDMA is used */
#define XTEA_DFT_RXBDS               32

/** Default number of Tx buffer descriptors if SGDMA is used */
#define XTEA_DFT_TXBDS               32

/** The maximum number of Tx fragments in a mblk chain the driver will
 *  map to descriptors. If the number of fragments is greater than this
 *  value, the driver will coalesce the fragments into a single buffer
 *  to conserve BD resources.
 */
#define XTEA_SG_MAX_FRAGS            8

/** The ratio of Rx buffers to use vs. Tx copy buffers for stack-to-driver
 *  copying. This constant is used only on HW configurations that do not allow
 *  misaligned data transfers to the transmit channel. See discussion about
 *  alignment requirements and buffer copying in xtemac_end_adapter.c for more
 *  information.
 */
#define XTEA_COPY_BUFS_RATIO         1

/** The number of buffers to reserve for use when coalescing mblk chains greater
 *  than XTEA_SG_MAX_FRAGS.
 */
#define XTEA_MAX_DEFRAG_BUFS         4

/** Define this constant to print out load-time information to the console that
 *  describes how many buffers and buffer descriptors were allocated. Undef this
 *  constant to prevent this information from being displayed.
 */
#define XTEA_PRINT_BUF_INIT_DATA

/** Maximum MTU. May be different than XTE_JUMBO_MTU */
#define XTEA_MAX_MTU     9014

/** Enable checksum offloading if supported by the stack */
#if (_WRS_VXWORKS_MAJOR >= 6)
#  define XTEA_CSUM_OFFLOAD
#endif

/** @name Custom ioctl commands
 */
/*@{*/

/** Set SG DMA packet threshold counts and waitbound timers */
#define XTEA_EIOCSPKTTW  129

/** Get SG DMA packet threshold counts and waitbound timers */
#define XTEA_EIOCGPKTTW  130

/** Set MAC-PHY data clock to match link speed */
#define XTEA_EIOCSMPSPD  133

/*@}*/


/**************************** Type Definitions *******************************/

/**
 * The END driver structure
 */
typedef struct
{
    /* VxWorks-specific */
    END_OBJ     VxEnd;              /**< END_OBJ for this device (must be 1st
                                         defined attribute) */
    int         VxUnit;             /**< VxWorks id */
    SEM_ID      SemConfig;          /**< Configuration semaphore */
    Xboolean    TxIsBlocked;        /**< The send channel is resource-blocked */
#ifdef XTEA_CSUM_OFFLOAD
    END_CAPABILITIES HwCaps;
#endif
    
    /* Buffer allocation data */
    Xuint32  BufSpaceAddr;          /**< Receive-buffer space addr */
    Xboolean IsBufMalloc;           /**< Was buffer memory malloc'd */

    /* Error data */
    END_ERR VxEndLastError;         /**< Last END error that occured */

    /* Processor-related */
    Xuint32     IntrId;             /**< Interrupt vector */
    CACHE_FUNCS cacheFuncs;         /**< cache function pointers */

    /* Misc */
    XTemac   Mac;                   /**< XTemac level 1 driver component */
    Xuint32  MaxFrameSz;            /**< Maximum frame size */

    /* SGDMA specific related */
    struct {
        Xuint32  BdSpaceAddr;       /**< Descriptor space address */
        Xboolean IsBdMalloc;        /**< Was buffer descriptor memory malloc'd */
    } SgDma;

    /* Receive buffer space managed by VxWorks netBufLib */
    struct {
        NET_POOL    NetPool;        /**< VxWorks Rx Network Pool */
        M_CL_CONFIG ClConfig;       /**< VxWorks Rx Cluster Config */
        unsigned    ClSize;         /**< Rx Cluster size */
        int         BufCnt;         /**< Total number of Rx buffers allocated */
    } RxBuf;

    /* Transmit buffer space managed locally. This buffer space is allocated
     * only when stack-to-driver buffer copying is required
     */
    struct {
        Xuint8  *BaseAddr;          /**< 1st Tx buffer address */
        Xuint8  *HighAddr;          /**< Last Tx buffer address */
        Xuint8  *NextBuf;           /**< Address of next available Tx buffer */
        Xuint32  BufSize;           /**< Tx buffer size */
        int      BufsAvailable;     /**< Number of Tx buffers left */
        int      BufCnt;            /**< Total number of Tx buffers allocated */
    } TxBuf;
    
} XTemacEnd;


/**
 * Structure used with custom IOCTL commands XEEA_EIOSPKTTW and
 * XEEA_EIOGPKTTW
 */
typedef struct
{
    Xuint16 RxPktWaitbound;     /**< Recv packet waitbound timer */
    Xuint16 TxPktWaitbound;     /**< Send packet threshold count */
    Xuint16 RxPktThresholdCnt;  /**< Recv packet threshold count */
    Xuint16 TxPktThresholdCnt;  /**< Send packet threshold count */
} XTemacEnd_SgPktIoctlData;


/***************** Macros (Inline Functions) Definitions *********************/


/************************** Function Prototypes ******************************/


END_OBJ *XTemacEndAdapter_Load(char *InitString, void *BspPtr);
XTemac  *XTemacEndAdapter_GetDriver(int UnitNumber);

#endif            /* end of protection macro */
