/* $Id: xuartns550_adapter.c,v 1.5 2003/02/14 21:41:35 robertm Exp $ */
/******************************************************************************
*
*       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*       SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*       FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*       FOR A PARTICULAR PURPOSE.
*
*       (c) Copyright 2002 Xilinx Inc.
*       All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xuartns550_adapter.c
*
* The implementation of the XUartNs550Adapter component, which is the adapter
* between VxWorks and the XUartNs550 driver.  This is intended for serial
* devices only.
*
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1.00b rmm  02/14/02 Fixed diab compiler warnings for Tornado 2.2
* </pre>
*
*****************************************************************************/

/***************************** Include Files ********************************/

/* VxWorks specific includes */

#include "vxWorks.h"
#include "intLib.h"
#include "iv.h"
#include "errnoLib.h"
#include "errno.h"
#include "sioLib.h"

/* Xilinx specific includes */
#include "xuartns550_adapter.h"

/************************** Constant Definitions ****************************/


/**************************** Type Definitions ******************************/


/***************** Macros (Inline Functions) Definitions ********************/


/************************** Function Prototypes *****************************/

/*
 * The VxWorks interface functions.  These can be static since they'll be
 * part of a function table passed to VxWorks.
 */
static int SendStartup(SIO_CHAN *ChanPtr);
static int PollRecv(SIO_CHAN *ChanPtr, char *InChar);
static int PollSend(SIO_CHAN *ChanPtr, char OutChar);
static int Ioctl(SIO_CHAN *ChanPtr, int Cmd, void *Data);
static int CallbackInstall(SIO_CHAN *ChanPtr, int CallBackType,
                           STATUS (*CallBack)(void*, ...), void *CallBackArg);


/*
 * Adapter internal functions
 */
static STATUS DummyCallback(void *CallBackRef, ...);
static void CallBackHandler(void *CallBackRef, Xuint32 Event,
                            unsigned int EventData);

static XStatus SetOpts(XUartNs550SioChannel *ChanPtr, UINT Options);
static XStatus SetMode(XUartNs550SioChannel *ChanPtr, UINT NewMode);
static XStatus Hup(XUartNs550SioChannel *ChanPtr);
static XStatus Open(XUartNs550SioChannel *ChanPtr);


/************************** Variable Definitions ****************************/

/*
 * The VxWorks function table for serial I/O drivers. Note that the order
 * of these function is critical since VxWorks uses offsets to access
 * the right function.
 */
static SIO_DRV_FUNCS SioDrvFuncs =
{
    Ioctl,
    SendStartup,
    (int (*)())CallbackInstall,
    PollRecv,
    PollSend
};


/*****************************************************************************
*
* FUNCTION:
*
* XUartNs550Adapter_DevInit
*
* DESCRIPTION:
*
* Initialize the adapter and the driver.  Set up the SIO_DRV_FUNCS function
* pointer table, and set the dummy callbacks for the interrupt-based
* get/put character routines (VxWorks will later set these callbacks
* to their appropriate functions).
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the XUartNs550SioChannel serial channel structure.
*
* InstanceNumber is the logical instance assigned to this channel.
*
* DeviceId is the unique device ID of the UART being initialized
*
* IntrId is the interrupt ID for the UART being initialized
*
* RETURN VALUE:
*
* XST_SUCCESS on successful initialization, or a driver-specific error
* code on failure.
*
* NOTES:
*
* None.
*
*****************************************************************************/
XStatus XUartNs550Adapter_DevInit(XUartNs550SioChannel *ChanPtr,
                                  Xuint16 DeviceId, int IntrId)
{
    XStatus Status;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(ChanPtr != NULL);

    /*
     * Disable interrupts for this device
     */
    ChanPtr->IntrId = IntrId;
    intDisable(ChanPtr->IntrId);

    /*
     * Set some channel values, including the function table pointer
     * and a dummy callback for now (until VxWorks sets appropriately)
     */
    ChanPtr->pDrvFuncs = &SioDrvFuncs;
    ChanPtr->getTxChar = DummyCallback;
    ChanPtr->putRcvChar = DummyCallback;
    ChanPtr->DeviceId = DeviceId;
    
    /*
     * Go ahead and create/initialize this driver and device
     */
    Status = XUartNs550_Initialize(&ChanPtr->UartNs550, ChanPtr->DeviceId);
    if (Status != XST_SUCCESS)
    {
        return Status;
    }

    /*
     * Set driver callback, used in interrupt mode. The callback reference
     * (the last argument) is the SIO channel pointer.
     */
    XUartNs550_SetHandler(&ChanPtr->UartNs550, CallBackHandler, ChanPtr);

    return XST_SUCCESS;
}


/*****************************************************************************
*
* FUNCTION:
*
* XUartNs550Adapter_DevInit2
*
* DESCRIPTION:
*
* Starts the device referred to by the device structure. This enables the
* device interrupt and connects the device's interrupt handler to the
* interrupt using the intConnect/intEnable VxWorks functions.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* RETURN VALUE:
*
* XST_SUCCESS on success, ERROR on failure.
*
* NOTES:
*
* None.
*
*****************************************************************************/
XStatus XUartNs550Adapter_DevInit2(XUartNs550SioChannel *ChanPtr)
{
    XStatus Status;
    Xuint16 Options;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(ChanPtr != NULL);

    /*
     * Connect the driver's interrupt handler to the interrupt and enable the
     * device's interrupt.
     */
    Options = XUartNs550_GetOptions(&ChanPtr->UartNs550);
    
    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550,
                                   Options | XUN_OPTION_DATA_INTR);
    if (Status != XST_SUCCESS)
    {
        return Status;
    }

    /*
     * Attach to interrupt driver and enable interupts
     */
    intConnect(INUM_TO_IVEC(ChanPtr->IntrId),
               (VOIDFUNCPTR)XUartNs550_InterruptHandler,
               (int)&ChanPtr->UartNs550);

    intEnable(ChanPtr->IntrId);

    return XST_SUCCESS;
}


/*****************************************************************************
*
* FUNCTION:
*
* SetOpts
*
* DESCRIPTION:
*
* Set the options of the device referred to by the device structure. This
* is the interface between the options available for the driver and the
* options expected by the RTOS. The VxWorks options are:
*
* - CSIZE is the number of data bits
* - STOPB is the number of stop bits
* - CLOCAL is the modem control settings
* - PARENB and PARODD are the parity settings
* - CREAD indicates the RTOS is ready for serial device processing
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* Options is a bit mask containing the requested VxWorks options
*
* RETURN VALUE:
*
* XST_SUCCESS on success, or driver-specific error on failure.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static XStatus SetOpts(XUartNs550SioChannel *ChanPtr, UINT Options)
{
    Xuint16 DeviceOptions;
    XUartNs550Format DataFormat;
    XStatus Status;
    int OldLevel;       /* current interrupt level Mask */

    /*
     * Disable interrupts for the following code
     */
    OldLevel = intLock();

    /*
     * Get the current data format, then step through the VxWorks
     * specific settings
     */
    XUartNs550_GetDataFormat(&ChanPtr->UartNs550, &DataFormat);

    /*
     * Set the size of the characters in bits
     */
    switch (Options & CSIZE)
    {
        case CS5:
            DataFormat.DataBits = XUN_FORMAT_5_BITS;
            break;

        case CS6:
            DataFormat.DataBits = XUN_FORMAT_6_BITS;
            break;

        case CS7:
            DataFormat.DataBits = XUN_FORMAT_7_BITS;
            break;

        default:
        case CS8:
            DataFormat.DataBits = XUN_FORMAT_8_BITS;
            break;
    }

    /*
     * Set the number of stop bits
     */
    if (Options & STOPB)
    {
        DataFormat.StopBits = XUN_FORMAT_2_STOP_BIT;
    }
    else
    {
        DataFormat.StopBits = XUN_FORMAT_1_STOP_BIT;
    }

    /*
     * Set the desired parity
     */
    switch (Options & (PARENB | PARODD))
    {
        case PARENB:
            DataFormat.Parity = XUN_FORMAT_EVEN_PARITY;
            break;

        case PARODD:
            DataFormat.Parity = XUN_FORMAT_ODD_PARITY;
            break;

        default:
        case 0:
            DataFormat.Parity = XUN_FORMAT_NO_PARITY;
            break;
    }

    /*
     * Update the driver data format
     */
    Status = XUartNs550_SetDataFormat(&ChanPtr->UartNs550, &DataFormat);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return Status;
    }

    /*
     * Now check for modem control
     */
    DeviceOptions = XUartNs550_GetOptions(&ChanPtr->UartNs550);

    if (!(Options & CLOCAL))
    {
        /*
         * clocal enables hardware flow control (DTR/DSR)
         */
        DeviceOptions |= (XUN_OPTION_ASSERT_RTS | XUN_OPTION_ASSERT_DTR |
                          XUN_OPTION_MODEM_INTR | XUN_OPTION_DATA_INTR);
    }
    else
    {
        DeviceOptions &= ~XUN_OPTION_MODEM_INTR;
    }

    /*
     * Update the driver configuration based on the settings above, and reset
     * and enable the FIFOs (a noop if the device is a 16450). We need to reset
     * the FIFOs then followup with a clearing of the reset bits.
     */
    DeviceOptions |= (XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO |
                      XUN_OPTION_FIFOS_ENABLE);

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return XST_FAILURE;
    }

    /* now clear the reset bits */
    
    DeviceOptions &= ~(XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO);

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return Status;
    }

    /*
     * If the OS is ready and we're in interrupt mode, go ahead and
     * enable the device interrupts.
     */
    if ((Options & CREAD) && (ChanPtr->ChannelMode == SIO_MODE_INT))
    {
        DeviceOptions |= XUN_OPTION_DATA_INTR;

        /*
         * update the driver configuration
         */
        Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
        if (Status != XST_SUCCESS)
        {
            intUnlock(OldLevel);
            return Status;
        }

    }

    /*
     * enable the interrupts
     */
    intUnlock(OldLevel);

    ChanPtr->Options = Options;

    return XST_SUCCESS;
}

/*****************************************************************************
*
* FUNCTION:
*
* SetMode
*
* DESCRIPTION:
*
* This driver supports both polled and interrupt modes and is capable of
* switching between modes dynamically.
*
* If interrupt mode is desired this routine enables the channel receiver and
* transmitter interrupts. If the modem control option is TRUE, the Tx interrupt
* is disabled if the CTS signal is FALSE. It is enabled otherwise.
*
* If polled mode is desired the device interrupts are disabled.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* RETURN VALUE:
*
* Returns XST_SUCCESS if the mode was set, XST_FAILURE if the mode is invalid,
* or a driver-specific error otherwise.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static XStatus SetMode(XUartNs550SioChannel *ChanPtr, UINT NewMode)
{
    Xuint16 DeviceOptions;
    XStatus Status;
    int OldLevel;       /* current interrupt level Mask */

    /*
     * check if NewMode is acceptable
     */
    if ((NewMode != SIO_MODE_POLL) && (NewMode != SIO_MODE_INT))
    {
        return XST_FAILURE;
    }

    /*
     * disable the interrupts
     */
    OldLevel = intLock();

    /*
     * get the current options
     */
    DeviceOptions = XUartNs550_GetOptions(&ChanPtr->UartNs550);

    if (NewMode == SIO_MODE_INT)
    {
        /* Enable appropriate interrupts */

        if (ChanPtr->Options & CLOCAL)
        {
            DeviceOptions |= XUN_OPTION_DATA_INTR;
        }
        else
        {
            /* hardware flow control */
            DeviceOptions |= (XUN_OPTION_ASSERT_RTS | XUN_OPTION_ASSERT_DTR);
        }
    }
    else
    {
        /*
         * polled mode, disable all interrupts
         */
        DeviceOptions &= ~XUN_OPTION_DATA_INTR;

    }

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return Status;
    }

    ChanPtr->ChannelMode = NewMode;

    /*
     * enable the interrupts
     */
    intUnlock(OldLevel);

    return XST_SUCCESS;
}


/*****************************************************************************
*
* FUNCTION:
*
* Hup
*
* DESCRIPTION:
*
* Resets the RTS and DTR signals and clears both the receiver and transmitter
* sections.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* RETURN VALUE:
*
* Returns XST_SUCCESS if successful hangup or a driver-specific error
* otherwise.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static XStatus Hup(XUartNs550SioChannel *ChanPtr)
{
    Xuint16 DeviceOptions;
    XStatus Status;
    int OldLevel;       /* current interrupt level Mask */

    /*
     * Disable the interrupts
     */
    OldLevel = intLock();
    
    /*
     * Get the current options
     */
    DeviceOptions = XUartNs550_GetOptions(&ChanPtr->UartNs550);

    /*
     * De-assert the RTS and DTR lines and reset the FIFOs
     */
    DeviceOptions &= ~(XUN_OPTION_ASSERT_RTS | XUN_OPTION_ASSERT_DTR);
    DeviceOptions |= (XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO);

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return Status;
    }

    /*
     * Return the FIFO's to the standard operating mode
     */
    DeviceOptions &= ~(XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO);
    DeviceOptions |= XUN_OPTION_FIFOS_ENABLE;

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
    if (Status != XST_SUCCESS)
    {
        intUnlock(OldLevel);
        return Status;
    }

    /*
     * Enable the interrupts
     */
    intUnlock(OldLevel);

    return XST_SUCCESS;
}


/*****************************************************************************
*
* FUNCTION:
*
* Open
*
* DESCRIPTION:
*
* Open a modem connection by setting the modem control lines (RTS, DTR) to
* TRUE if not already set. Also clear the receiver, transmitter and enables
* the fifo.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* RETURN VALUE:
*
* Returns XST_SUCCESS if successful hangup or a driver-specific error
* otherwise.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static XStatus Open(XUartNs550SioChannel *ChanPtr)
{
    Xuint16 DeviceOptions;
 
    /*
     * Get the current options
     */
    DeviceOptions = XUartNs550_GetOptions(&ChanPtr->UartNs550);

    /*
     * Only do work if the modem control lines are not already asserted.
     * Assert the DTR and RTS lines to allow data to flow
     */
    if ((DeviceOptions & (XUN_OPTION_ASSERT_RTS | XUN_OPTION_ASSERT_DTR)) == 0)
    {
        XStatus Status;
        int OldLevel;       /* current interrupt level Mask */

        OldLevel = intLock();

        /*
         * Assert RTS and DTR and reset the FIFOs
         */
        DeviceOptions |= (XUN_OPTION_ASSERT_RTS | XUN_OPTION_ASSERT_DTR |
                          XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO);

        Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
        if (Status != XST_SUCCESS)
        {
            intUnlock(OldLevel);
            return Status;
        }

        /*
         * Return the FIFO's to the standard operating mode
         */
        DeviceOptions &= ~(XUN_OPTION_RESET_TX_FIFO | XUN_OPTION_RESET_RX_FIFO);
        DeviceOptions |= XUN_OPTION_FIFOS_ENABLE;

        Status = XUartNs550_SetOptions(&ChanPtr->UartNs550, DeviceOptions);
        if (Status != XST_SUCCESS)
        {
            intUnlock(OldLevel);
            return Status;

        }

        /*
         * Enable the interrupts
         */
        intUnlock(OldLevel);
    }

    return XST_SUCCESS;
}

/*****************************************************************************
*
* FUNCTION:
*
* CallbackInstall
*
* DESCRIPTION:
*
* This driver allows interrupt callbacks for transmitting characters
* and receiving characters. In general, drivers may support other
* types of callbacks too.  This function is called by VxWorks to set the
* callbacks to appropriate routines.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure
*
* CallBackType is either SIO_CALLBACK_GET_TX_CHAR or SIO_CALLBACK_PUT_RCV_CHAR
*
* CallBack is the callback function
*
* CallBackArg is the argument that will be passed to the callback function
* when it is invoked.
*
* RETURN VALUE:
*
* OK on success, or ENOSYS for an unsupported callback type.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static int CallbackInstall(SIO_CHAN *SioChanPtr, int CallBackType,
                           STATUS (*CallBack)(void*, ...), void *CallBackArg)
{
    XUartNs550SioChannel *ChanPtr;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(SioChanPtr != NULL);
    XASSERT_NONVOID(CallBack != NULL);
    XASSERT_NONVOID(CallBackArg != NULL);


    ChanPtr = (XUartNs550SioChannel *)SioChanPtr;

    /*
     * Select the callback type to update the proper structure elements
     */
    switch (CallBackType)
    {
        case SIO_CALLBACK_GET_TX_CHAR:
            ChanPtr->getTxChar = CallBack;
            ChanPtr->getTxArg = CallBackArg;
            break;

        case SIO_CALLBACK_PUT_RCV_CHAR:
            ChanPtr->putRcvChar = CallBack;
            ChanPtr->putRcvArg = CallBackArg;
            break;

        default:
            /*
             * Unrecognized callback type
             */
            return ENOSYS;
    }

    return OK;
}


/*****************************************************************************
*
* FUNCTION:
*
* SendStartup
*
* DESCRIPTION:
*
* Start the interrupt transmitter.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure.
*
* RETURN VALUE:
*
* OK on success, ENOSYS if the device is polled-only, or EIO on hardware error.
*
* NOTES:
*
*****************************************************************************/
static int SendStartup(SIO_CHAN *SioChanPtr)
{
    XStatus Status;
    Xuint16 DeviceOptions;
    XUartNs550SioChannel *ChanPtr;
    Xuint8 DataChar;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(SioChanPtr != NULL);

    ChanPtr = (XUartNs550SioChannel *)SioChanPtr;
    
    if (ChanPtr->ChannelMode == SIO_MODE_POLL)
    {
        return ENOSYS;
    }

    /*
     * Enable the interrupt
     */
    DeviceOptions = XUartNs550_GetOptions(&ChanPtr->UartNs550);

    Status = XUartNs550_SetOptions(&ChanPtr->UartNs550,
                                   DeviceOptions | XUN_OPTION_DATA_INTR);
    if (Status != XST_SUCCESS)
    {
        return ENOSYS;
    }

    /*
     * Kick the transmitter to get the interrupt to occur for the first time
     */
    if ((*ChanPtr->getTxChar)(ChanPtr->getTxArg, &DataChar) != ERROR)
    {
        (void)XUartNs550_Send(&ChanPtr->UartNs550, &DataChar, 1);
    }

    return OK;
}


/*****************************************************************************
*
* FUNCTION:
*
* PollSend
*
* DESCRIPTION:
*
* Send a character in polled mode.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure
*
* OutChar is the character to send
*
* RETURN VALUE:
*
* OK if a character is sent, EIO on device error, EAGAIN if the output buffer
* if full. ENOSYS if the device is interrupt-only.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static int PollSend(SIO_CHAN *SioChanPtr, char OutChar)
{
    Xuint8 NumBytes;
    XUartNs550SioChannel *ChanPtr;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(SioChanPtr != NULL);

    ChanPtr = (XUartNs550SioChannel *)SioChanPtr;

    /*
     * If the device is in interrupt mode, invalid request
     */
    if (ChanPtr->ChannelMode == SIO_MODE_INT)
    {
        return ENOSYS;
    }

    /*
     * Send the byte
     */
    NumBytes = XUartNs550_Send(&ChanPtr->UartNs550, &OutChar, 1);

    if (NumBytes == 0)
    {
        return EAGAIN;
    }

    return OK;
}

/*****************************************************************************
*
* FUNCTION:
*
* PollRecv
*
* DESCRIPTION:
*
* Poll the device for a received character.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure
*
* InChar is a pointer to a character in which the received character will
* be placed
*
* RETURN VALUE:
*
* OK if a character arrived, EIO on device error, EAGAIN if the input buffer
* is empty, ENOSYS if the device is interrupt-only.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static int PollRecv(SIO_CHAN *SioChanPtr, char *InChar)
{
    Xuint8 NumBytes;
    XUartNs550SioChannel *ChanPtr;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(SioChanPtr != NULL);
    XASSERT_NONVOID(InChar != NULL);

    ChanPtr = (XUartNs550SioChannel *) SioChanPtr;

    /*
     * If the device is in interrupt mode, invalid request
     */
    if (ChanPtr->ChannelMode == SIO_MODE_INT)
    {
        return ENOSYS;
    }

    /*
     * Receive one byte
     */
    NumBytes = XUartNs550_Recv(&ChanPtr->UartNs550, InChar, 1);

    if (NumBytes == 0)
    {
        return EAGAIN;
    }

    return OK;
}


/*****************************************************************************
*
* FUNCTION:
*
* Ioctl
*
* DESCRIPTION:
*
* Device control. This routine handles the IOCTL messages from the user.
* It supports commands to get/set baud rate, mode (INT,POLL), hardware
* options (parity, number of data bits) and modem control (RTS/CTS and
* DTR/DSR handshakes). The ioctl commands SIO_HUP and SIO_OPEN are used
* to implement the HUPCL (hang up on last close) function.
*
* ARGUMENTS:
*
* ChanPtr is a pointer to the serial channel structure
*
* Cmd is the action requested
*
* Data is the data passed in for some commands or returned for other commands
*
* RETURN VALUE:
*
* OK on success, ENOSYS on unsupported request, EIO on failed request.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static int Ioctl(SIO_CHAN *SioChanPtr, int Cmd, void *Data)
{
    XStatus Result;
    STATUS Status;
    XUartNs550SioChannel *ChanPtr;
    XUartNs550Format DataFormat;

    /*
     * Assert validates the input arguments
     */
    XASSERT_NONVOID(SioChanPtr != NULL);

    ChanPtr = (XUartNs550SioChannel *) SioChanPtr;

    Status = OK;

    /*
     * Step through the valid commands
     */
    switch (Cmd)
    {
        case SIO_BAUD_SET:

            if ((Xuint32)Data < XUN_NS16550_MIN_RATE ||
                (Xuint32)Data > XUN_NS16550_MAX_RATE)
            {
                Status = EIO;       /* baud rate out of range */
            }
            else
            {
                XUartNs550_GetDataFormat(&ChanPtr->UartNs550, &DataFormat);

                DataFormat.BaudRate = (Xuint32)Data;

                Result =
                    XUartNs550_SetDataFormat(&ChanPtr->UartNs550, &DataFormat);
                if (Result != XST_SUCCESS)
                {
                    Status = EIO;       /* baud rate out of range/too much error */
                }
            }
            break;

        case SIO_BAUD_GET:

            XUartNs550_GetDataFormat(&ChanPtr->UartNs550, &DataFormat);

            *(int *)Data = DataFormat.BaudRate;
            break;

        case SIO_MODE_SET:

            Result = SetMode(ChanPtr, (UINT)Data);
            if (Result != XST_SUCCESS)
            {
                Status = EIO;
            }
            break;

        case SIO_MODE_GET:

            *(int *)Data = ChanPtr->ChannelMode;
            break;

        case SIO_AVAIL_MODES_GET:

            *(int *)Data = (SIO_MODE_INT | SIO_MODE_POLL);
            break;

        case SIO_HW_OPTS_SET:

            Result = SetOpts(ChanPtr, (UINT)Data);
            if (Result != XST_SUCCESS)
            {
                Status = EIO;
            }
            break;

        case SIO_HW_OPTS_GET:

            *(int *)Data = ChanPtr->Options;
            break;

        case SIO_HUP:

            Result = Hup(ChanPtr);
            if (Result != XST_SUCCESS)
            {
                Status = EIO;
            }
            break;

        case SIO_OPEN:

            Result = Open(ChanPtr);
            if (Result != XST_SUCCESS)
            {
                Status = EIO;
            }
            break;

        default:
            return ENOSYS;
    }

    return Status;
}


/*****************************************************************************
*
* FUNCTION:
*
* CallBackHandler
*
* DESCRIPTION:
*
* Callback function (called from driver) to handle interrupts
*
* ARGUMENTS:
*
* CallBackRef is the callback reference passed back from the driver.  It is
* a pointer to the serial channel structure, indicating which channel just
* received data.
*
* Event is which interrupt source caused callback
*
* EventData is event specific data
*
* RETURN VALUE:
*
* None.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static void CallBackHandler(void *CallBackRef, Xuint32 Event,
                             unsigned int EventData)
{
    Xuint8 DataChar;
    Xuint8 NumBytes;
    XUartNs550SioChannel *ChanPtr;

    ChanPtr = (XUartNs550SioChannel *) CallBackRef;

    switch (Event)
    {
        case XUN_EVENT_RECV_TIMEOUT:
        case XUN_EVENT_RECV_ERROR:
        case XUN_EVENT_RECV_DATA:
            /*
             * There is probably data to be received, so receive one
             * character then pass it up to VxWorks.
             */
            NumBytes =
                XUartNs550_Recv(&ChanPtr->UartNs550, &DataChar, 1);

            if (NumBytes != 0)
            {
                ((*ChanPtr->putRcvChar)(ChanPtr->putRcvArg, DataChar));
            }
            break;

        case XUN_EVENT_SENT_DATA:
            /*
             * The transmitter is ready for more data, so check with VxWorks
             * to see if it has something more to send.
             */
            if ((*ChanPtr->getTxChar)(ChanPtr->getTxArg, &DataChar) != ERROR)
            {
                (void)XUartNs550_Send(&ChanPtr->UartNs550, &DataChar, 1);
            }
            break;

        case XUN_EVENT_MODEM:
        default:
            break;
    }
}


/*****************************************************************************
*
* FUNCTION:
*
* DummyCallback
*
* DESCRIPTION:
*
* Dummy callback function for get/put character callbacks.  These callbacks
* are part of the serial channel structure and are set by VxWorks.  But, a
* stub is used until VxWorks sets the callbacks to avoid chaos.
*
* ARGUMENTS:
*
* None.
*
* RETURN VALUE:
*
* Always returns ERROR.
*
* NOTES:
*
* None.
*
*****************************************************************************/
static STATUS DummyCallback(void *CallbackRef, ...)
{
    return ERROR;
}
