/* sysInterrupt.c - System interrupt hardware setup */

/*******************************************************************************
   XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
   AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
   SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
   OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
   APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
   THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
   AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
   FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
   WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
   IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
   REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
   INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE.

   (c) Copyright 2003 Xilinx Inc.
   All rights reserved.
*******************************************************************************/

/*
modification history
--------------------
01a,18feb03,rmm  Ported from ML300
*/

/*
DESCRIPTION
This library contains board-specific routines for coordination and
initialization of system interrupt hardware for PPC processors.

In this design example, two Xilinx interrupt controllers are utilized. One is
connected to the external interrupt signal and the other to the critical
interrupt signal. The XIntc device driver is utilized to control the hardware.

Two XIntc instances are initialized, one for each HW device. The instance
associated with the external interrupt signal is hooked into VxWorks' intLib
library, so calls to intConnect(), intEnable(), intDisable() will call
functions here which will in turn invoke the XIntc driver to perform the
intended operation. Mirror functions sysIntCritConnect(), sysIntCritEnable(),
sysIntCritDisable() will handle operations to sources on the critical
interrupt handler.

*/

#include "vxWorks.h"
#include "intLib.h"
#include "esf.h"
#include "config.h"

/*
 * Include header file for each interrupt controller in the system.
 */
#include "xstatus.h"

#ifdef INCLUDE_XINTC
#  include "xintc.h"
#endif

/*
 * Defines
 */
#define INT_LEVELS 32

/*
 * Local functions
 */
static STATUS intcConnect(VOIDFUNCPTR *vector, VOIDFUNCPTR routine,
                            int parameter);
static STATUS intcEnable(int intLevel);
static STATUS intcDisable(int intLevel);
static void sysExternalISR(void);
static void sysCriticalISR(void);

/*
 * Externals
 *
 * _func functions are VxWorks hooks for functions intConnect, intEnable,
 * and intDisable.
 */
extern STATUS (*_func_intConnectRtn) (VOIDFUNCPTR *, VOIDFUNCPTR, int);
extern STATUS (*_func_intEnableRtn) (int);
extern STATUS (*_func_intDisableRtn) (int);

/*
 * Global  data
 */
#ifdef INCLUDE_XINTC
XIntc sysXIntcExt;   /* Xintc Instance hooked to external interrupt pin */

/******************************************************************************
*
* sysIntInit - Initialize the system interrupt HW & SW interface.
*
* This routine initializes interrupt controller HW and its interface to VxWorks
* It is called from sysHwInit() with interrupts locked. There may be two intc
* controllers, one connected to the external interrupt signal and the other to
* the critical interrupt signal. Two XIntc instances from the CSP are
* instantiated with the first one hooked into the VxWorks intLib library. The
* second one requires the user to connect/enable/disable interrupts.
*
* RETURNS: OK if successful, ERROR otherwise
*
* SEE ALSO: sysHwInit()
*/
STATUS sysIntInit(void)
{
    /*
     * Initialize the XIntc instance connected to the external interrupt pin.
     */
    if (XIntc_Initialize(&sysXIntcExt, XPAR_OPB_INTC_0_DEVICE_ID) != XST_SUCCESS)
    {
        BOOT_LOG_ERROR();
        return(ERROR);
    }

    if (XIntc_Start(&sysXIntcExt, XIN_REAL_MODE) != XST_SUCCESS)
    {
        BOOT_LOG_ERROR();
        return(ERROR);
    }

    /*
     * Setup external interrupt hook functions
     */
    _func_intConnectRtn = intcConnect;
    _func_intEnableRtn = intcEnable;
    _func_intDisableRtn = intcDisable;

    /*
     * Attach to external interrupt vectors
     */
    excIntConnect((VOIDFUNCPTR *)_EXC_OFF_INTR, (VOIDFUNCPTR)sysExternalISR);
    excIntCrtConnect((VOIDFUNCPTR *)_EXC_OFF_CRTL, (VOIDFUNCPTR)sysCriticalISR);
    return(OK);
}

/******************************************************************************
*
* intcConnect - Connect an interrupt
*
* This routine connects device interrupts. This function is hooked into the
* vxWorks intConnect() library function call. Only interrupts sourced from
* the XIntc device on the external interrupt pin may be connected with this
* routine.
*
* RETURNS: OK if successful, ERROR if vector is out of range or an error
* occurred connecting routine in XIntc component.
*
*/
static STATUS intcConnect
    (
    VOIDFUNCPTR* vector,   /* Interrupt ID vector */
    VOIDFUNCPTR routine,   /* Routine to connect */
    int parameter          /* parameter to routine */
    )
{
    int intLevel;
    XStatus status;

    /*
     * Check for errors in arguments
     */
    if ((unsigned)vector > INT_LEVELS)
    {
        return(ERROR);
    }

    /*
     * Disable interrupts while updating the component driver
     */
    intLevel = intLock();
    XIntc_Disconnect(&sysXIntcExt, IVEC_TO_INUM(vector));
    status = XIntc_Connect(&sysXIntcExt, IVEC_TO_INUM(vector),
                           (XInterruptHandler)routine, (void*)parameter);

    intUnlock(intLevel);

    if (status != XST_SUCCESS)
    {
        return(ERROR);
    }
    return(OK);
}

/******************************************************************************
*
* intcEnable - Enable interrupts for a vector
*
* This function is hooked into the VxWorks intEnable() library routine.
* Only sources connected to the XIntc instance on the external interrupt pin
* are affected.
*
* RETURNS: OK on success, ERROR otherwise
*
*/
static STATUS intcEnable
    (
    int intLevel    /* Vector to enable */
    )
{
    /*
     * Check for errors in arguments
     */
    if ((unsigned)intLevel > INT_LEVELS)
    {
        return(ERROR);
    }

    /*
     * Component does the actual connect
     */
    XIntc_Enable(&sysXIntcExt, (Xuint8)intLevel);
    return(OK);
}

/******************************************************************************
*
* intcDisable - Disable interrupts for a vector
*
* This function is hooked into the VxWorks intDisable() library routine.
* Only sources connected to the XIntc instance on the external interrupt pin
* are affected.
*
* RETURNS: OK on success, ERROR otherwise
*
*/
static STATUS intcDisable
    (
    int intLevel   /* Vector to disable */
    )
{
    /*
     * Check for errors in arguments
     */
    if ((unsigned)intLevel > INT_LEVELS)
    {
        return(ERROR);
    }

    /*
     * Component does the actual disconnect
     */
    XIntc_Disable(&sysXIntcExt, (Xuint8)intLevel);
    return(OK);
}

/******************************************************************************
*
* sysExternalISR - Handler for external interrupt exceptions
*
* External interrupt handler. Do whatever is necessary to handle critical
* interrupts here.
*
* RETURNS: N/A
*
*/
static void sysExternalISR(void)
{
    XIntc_InterruptHandler(&sysXIntcExt);
}

/******************************************************************************
*
* sysCriticalISR - Handler for critical interrupt exceptions
*
* System critical error handler. Do whatever is necessary to handle critical
* interrupts here.
*
* RETURNS: N/A
*
*/
static void sysCriticalISR(void)
{
}


/*
 * The code that follows supports the scenario where there are no interrupt
 * controllers. A single device is wired to the external interrupt signal
 * of the processor. This may be the case during system development.
 */
#else  /* INCLUDE_XINTC */

/*
 * Local data
 */
static FUNCPTR extISR;         /* Device interrupt handler on external vector */
static int extISR_parameter;   /* Parameter to handler */
static int extLocked = TRUE;   /* Interrupt locked flag */

static FUNCPTR critISR;        /* Device interrupt handler on critical vector */
static int critISR_parameter;  /* Parameter to handler */
static int critLocked = TRUE;  /* Interrupt locked flag */

/******************************************************************************
*
* sysIntInit - Initialize the system interrupt HW & SW interface for
* systems without an external interrupt controller.
*
* This function is called from sysHwInit() with interrupts locked.
*
* RETURNS: OK
*
* SEE ALSO: sysHwInit()
*/
STATUS sysIntInit(void)
{
    /*
     * Connect routies
     */
    excIntConnect((VOIDFUNCPTR *)_EXC_OFF_INTR, (VOIDFUNCPTR)sysExternalISR);
    excIntCrtConnect((VOIDFUNCPTR *)_EXC_OFF_CRTL, (VOIDFUNCPTR)sysCriticalISR);

    /*
     * Setup hook functions
     */
    _func_intConnectRtn = intcConnect;
    _func_intEnableRtn = intcEnable;
    _func_intDisableRtn = intcDisable;

    /*
     * Null out extISR until intConnect is called
     */
    extISR = (FUNCPTR)0;
    extLocked = TRUE;

    critISR = (FUNCPTR)0;
    critLocked = TRUE;
    return(OK);
}

/******************************************************************************
*
* intcConnect - Connect an interrupt
*
* This routine connects device interrupts directly to the external interrupt
* vector.
*
* The vector parameter is ignored. The routine and parameter parameters
* replace ones from a previous call to this routine.
*
* RETURNS: OK
*
* SEE ALSO: sysHwInit2()
*/
static STATUS intcConnect
    (
    VOIDFUNCPTR* vector,   /* Interrupt ID vector */
    VOIDFUNCPTR routine,   /* Routine to connect */
    int parameter          /* parameter to routine */
    )
{
    int intLevel;

    intLevel = intLock();

    /*
     * It is up to the user as to which vector is reserved to external and
     * critical interrupts. Default is to use external for all vectors.
     */
    extISR = (FUNCPTR)routine;
    extISR_parameter = parameter;

    intUnlock(intLevel);
    return(OK);
}

/******************************************************************************
*
* intcEnable - Enable interrupts for a vector which in this case enables
* external interrupts and optionally critical interrupts.
*
* RETURNS: OK
*
*/
static STATUS intcEnable
    (
    int intLevel   /* Vector to enable */
    )
{
    int level;

    /*
     * It is up to the user as to which vector is reserved to external and
     * critical interrupts. Default is to use external for all vectors.
     */
    level = intLock();
    extLocked = FALSE;
    intUnlock(level);
    return(OK);
}

/******************************************************************************
*
* intcDisable - Disable interrupts for a vector which in this case disables
* external interrupts and optionally critical interrupts.
*
* RETURNS: OK
*
*/
static STATUS intcDisable
    (
    int intLevel   /* Vector to disable */
    )
{
    int level;

    /*
     * Can't just switch off async interrupts MSR(EE) because that turns off
     * PIT and FIT interrupts which is not what the caller of intDisable expects
     * to happen. We can use a global to check in the ISR. If it is locked,
     * then just don't call the handler.
     *
     * It is up to the user as to which vector is reserved to external and
     * critical interrupts. Default is to use external for all vectors.
     */
    level = intLock();
    extLocked = TRUE;
    intUnlock(level);
    return(OK);
}

/******************************************************************************
*
* sysExternalISR - Handler for non-critical interrupt exceptions
*
* System non-critical error handler. Do whatever is necessary to handle non-
* critical interrupts here.
*
* RETURNS: N/A
*
*/
static void sysExternalISR(void)
{
    if (extISR && !extLocked)
    {
        (extISR)(extISR_parameter);
    }
}

/******************************************************************************
*
* sysCriticalISR - Handler for critical interrupt exceptions
*
* System critical error handler. Do whatever is necessary to handle critical
* interrupts here.
*
* RETURNS: N/A
*
*/
static void sysCriticalISR(void)
{
    if (critISR && !extLocked)
    {
        (critISR)(critISR_parameter);
    }
}

#endif /* INCLUDE_XINTC */

