/* sysSystemAce.c - System ACE board level routines */

/* NOTE - CR 193753 pre-release */

/*******************************************************************************
   XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
   AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
   SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
   OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
   APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
   THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
   AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
   FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
   WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
   IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
   REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
   INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE.

   (c) Copyright 2003 Xilinx Inc.
   All rights reserved.
*******************************************************************************/

/*

TODO - Start a new revision history

modification history
--------------------
01a,18feb03,rmm  Ported from ML300 version 01b
01b,02sep04,rmm  CR 193753. Fixed multiple partition support.
01c,12jul06,wsy  incorporated XBD changes(initiated from windriver) in.
*/

/*
DESCRIPTION
This library contains board-specific routines to initialize the Xilinx
System ACE controller. System ACE is a very flexible storage solution
primarily used to contain bitstreams within a DOSFs filesystem on a Compact
Flash storage medium.

System ACE JTAG interface:

The bitstreams files (created by the Xilinx tool chain) can contain FPGA IP
data as well as a memory image. The memory image can be any type of data
including a vxWorks bootrom image. The bitstreams are stored in a DOS file
system. Up to eight different bitstream configurations may be present. Any one
of these bitstreams may be selected to be loaded either programatically or by
the rotary switch on the main board. When the system is reset, System ACE will
load the FPGA with the IP bitstream then if there is a memory image it will be
loaded into RAM and the processor's program counter set to the entry point.


System ACE MPU interface:

The MPU interface allows the System ACE device to be accessed under system
software control. It allows software to read/write blocks of data to the
compact flash device and allows software initiated system resets. The Xilinx
component driver exports a block device interface which allows VxWorks to
view the compact flash as an external disk.


Bitstream file requirements:

Even though System ACE maintains a DOS file system, some files must be
located in certain directories.

    ACE Files
    ---------
    These are specially formatted files created from FPGA bitstream files. The
    files are created by the Xilinx tool chain and have a file extension of
    .ace. These files may contain the FPGA's IP data and/or a memory image to
    be downloaded to RAM.

    xilinx.sys
    ----------
    This file located in the root directory specifies the directories
    where the bitstream files may be located. Up to eight directories can be
    specified. Each directory may contain a single ACE file. If xilinx.sys is
    not present, then a single ACE file must be in the root directory. The
    following example illustrates a xinlinx.sys and the associated directory
    structure:

       xilinx.sys contents          required directory structure
       -------------------          ----------------------------
       dir = Rev0                     /Rev0
       cfgaddr0 = vxboot                |---vxboot
       cfgaddr1 = lastGood              |---lastGood
       cfgaddr2 = vxapp1                |---vxapp1
       cfgaddr3 = vxapp2                |---vxapp2
       cfgaddr4 = debug1                |---debug1
       cfgaddr5 = experiment1           |---experiment1
       cfgaddr6 = experiment2           |---experiment2
       cfgaddr7 = experiment3           |---experiment3


System ACE as a boot device:

   Two modes of bootup can be utilized with System ACE, JTAG and MPU (see
   previous paragraphs). Which mode is utilized depends on the bootline.

   The bootline should be specified as follows:

      sysace=P(X,Y):FILE [rest of bootline...]

   where "FILE" can take one of two forms. To boot using the JTAG interface,
   specify "cfgaddrZ" where "Z" is either not specified or is a number
   between 0..7. If a number then the matching configuration is loaded (see
   previous discussion on xilinx.sys). If "Z" is not specified, then the
   default configuration is loaded. To boot using the MPU interface "FILE"
   specifies a regular filename in the form of /mountpoint/path/to/file.

   "P" is the partition number to boot from when using the MPU interface. When
    specified as "0", the compact flash device is mounted as if it were a
    huge floppy disk (which do not have partition tables).

   "X" is the logical device number of the system ACE in use (normally 0).
   "Y" is the proc number.

   Examples:

      1. sysace=1(0,0):cfgaddr4
      2. sysace=1(0,0):cfgaddr
      3. sysace=1(0,0):/cf0/vxWorks
      4. sysace=0(0,0):/cf0/vxWorks

      The first example boots the .ACE JTAG file from the directory specified
      by cfgaddr4 (see xilinx.sys).

      The second example boots the .ACE JTAG file from the default directory.
      The default directory is selected by jumpers, switches, or hardwired
      signals to the System ACE controller.

      The third example boots the ELF file vxWorks located in the root directory
      of the first partition.

      The fourth example boots the ELF file vxWorks located in the root
      directory. The compact flash device is assumed to not contain a
      partition table.


INCLUDE FILES: sysLibExtra.h
*/

#ifndef  __INCsysSystemAce
#define  __INCsysSystemAce

#include "stdio.h"

#ifdef INCLUDE_XSYSACE
#include "xsysace_l.h"
#endif

#ifdef INCLUDE_XSYSACE_VXWORKS5_4
#include "dosFsLib.h"
#include "xbdBlkDev.h"
#include "xsysaceblkadapter.h"
BLK_DEV *pAce;
/* external variables */
extern void* usrFdiskPartRead;  /* Library module provided by VxWorks */
#endif

#define SYSACE_CFGADDR_DEFAULT 100

/* local variables */
static unsigned sysAceCfgAddr = SYSACE_CFGADDR_DEFAULT;

/*******************************************************************************
*
* sysSystemAceInit - Initialize System ACE
*
* If INCLUDE_XSYSACE_INSTALL_RESET_VEC is defined, then reset code will be
* installed at the CPU's reset vector. Normally a design will have some sort
* of reset code at this address, but in case it doesn't the capability to
* install it exists here. The code installed will initiate a SystemAce reset
* on the default configuration address even if sysSystemAceSetRebootAddr()
* had been called previously.
*
* RETURNS: OK
*
*/
STATUS sysSystemAceInit(void)
{
    return(OK);
}


/*******************************************************************************
*
* sysSystemAceInitFS - Initialize dosFS 2.0 file system support.
*
* Make calls to various DOS Filesystem 2.0 initialization functions to get the
* system ready to mount the compact flash device.
*
* RETURNS: N/A
*
*/
void sysSystemAceInitFS(void)
{
    /*
     * Function no longer required for Tornado 2.2
     * Kept as a placeholder for future system ace file system issues and
     * to remain backward compatibility with Tornado 2.0 apps
     */
}


/*******************************************************************************
*
* sysSystemAceMount - Mount a DOS file system in a System ACE controller.
*
* The <mountpoint> parameter is an arbitrary string labeling the device. Once
* mounted, refer to this mountpoint in all file accesses.
*
* The <partition> parameter specifies the partition to mount. If "0" is
* specified then the boot device is assumed to not contain a partition table
* (i.e. it is treated like a floppy disk)
*
* NOTE: Before calling this routine, be sure to initialize the DOS file system
* with a call to sysSystemAceInitFS().
*
* RETURNS: OK or ERROR.
*
* SEE ALSO:
*  sysSystemAceInitFS()
* .pG "I/O System, Local File Systems"
*/
STATUS sysSystemAceMount
    (
    unsigned short aceDeviceId,     /* System ACE device ID */
    char* mountPoint,    /* Mount point label */
    int partition        /* Partition to mount */
    )
{
#if defined(INCLUDE_XSYSACE) && defined(INCLUDE_XSYSACE_VXWORKS5_4) && ((defined(INCLUDE_XBD_BLK_DEV) && defined(INCLUDE_XBD_PART_LIB)) )
    static int initialized = 0;
    int i;
    char fsmName[16];
    device_t dev;
    unsigned numBlocks;

    /*
     * If not already initialized, initialize the System ACE/CompactFlash
     * as a block device.
     */
    if (!initialized)
    {
        if (XSysAceBlkAdapter_Init(aceDeviceId, 1) != OK)
        {
            return(ERROR);
        }

        /* get number of blocks. */
        numBlocks = XSysAceBlkAdapter_GetTotalBlocks(aceDeviceId);

        /* create the block device */
        pAce = XSysAceBlkAdapter_DevCreate(aceDeviceId, 0,
                                           numBlocks, 0);

        if (pAce == NULL)
        {
            printErr("error creating System ACE block device.\n");
            return(ERROR);
        }

        initialized = 1;
    }

    sprintf(fsmName, "%s:0", mountPoint);

    fsmNameInstall(fsmName, mountPoint);
    for (i = 0; i < SYSACE_MAX_PARTITIONS; i++)
    {
        sprintf(fsmName, "%s:%d", mountPoint, i+1);
        fsmNameInstall(fsmName, mountPoint);
    }

    dev = xbdBlkDevCreateSync(pAce, mountPoint);
    if (dev == NULLDEV)
    {
        printErr("error creating XBD\n");
        return(ERROR);
    }

    return (OK);

    /* Error if no system ace support */
#else
    return(ERROR);
#endif
}


/*******************************************************************************
*
* sysSystemAceSetJtagConfigAddr - Set the System ACE's JTAG reboot address
*
* This routine will note the reset address from which System ACE will reboot
* from should sysToMonitor() be invoked().
*
* The <configAddr> parameter selects the directory to find the .ACE file to
* boot. Valid ranges are 0..7 or -1 to select the default configuration.
*
* RETURNS: OK or ERROR if configAddr is out of range.
*
*/
STATUS sysSystemAceSetJtagConfigAddr
    (
    unsigned configAddr  /* Configuration address */
    )
{
#ifdef INCLUDE_XSYSACE
    /* default ? */
    if (configAddr == -1)
    {
        sysAceCfgAddr = SYSACE_CFGADDR_DEFAULT;
        return(OK);
    }

    /* out of range */
    if (configAddr > 7)
    {
        return(ERROR);
    }

    /* within range */
    sysAceCfgAddr = configAddr;
    return(OK);

    /* Error if no system ace support */
#else
    return(ERROR);
#endif
}


/*******************************************************************************
*
* sysSystemAceJtagReboot - Initiate System ACE JTAG reboot.
*
* Trigger System ACE to reboot using its JTAG downloader.
*
* RETURNS: Does not return if reboot successfully initiated. ERROR is
* returned otherwise.
*
*/

STATUS sysSystemAceJtagReboot(
    unsigned long baseAddr     /* Base address of System ACE device */
)
{
#ifdef INCLUDE_XSYSACE
    int delay = 10;

    /*
     * If system ace is connected as a 16 bit device, then we must tell the
     * device so that it presents 16 bit registers.  Level 1 driver does this
     * for us, but this function uses level 0 driver.
     */
#if (XPAR_XSYSACE_MEM_WIDTH == 16)
    XSysAce_RegWrite16(baseAddr + XSA_BMR_OFFSET, XSA_BMR_16BIT_MASK);
#endif

    /* Set the system ace config address to the default value */
    if (sysAceCfgAddr == SYSACE_CFGADDR_DEFAULT)
    {
        XSysAce_mAndControlReg(baseAddr, ~XSA_CR_FORCECFGADDR_MASK);
    }

    /* Set the system ace config address to the specified value */
    else
    {
        XSysAce_mSetCfgAddr(baseAddr, sysAceCfgAddr);
    }

    /* Now reset the system */
    XSysAce_mOrControlReg(baseAddr, (XSA_CR_CFGMODE_MASK | XSA_CR_CFGRESET_MASK));
    XSysAce_mAndControlReg(baseAddr, ~XSA_CR_CFGRESET_MASK);

    /* By the time this delay retuns, reboot should have been initiated */
    while(delay--)
    {
        sysMsDelay(1000);
    }

#endif  /* INCLUDE_XSYSACE */

    /*
     * Return error in case reboot didn't happen or there is no System
     * ace support
     */
    return(ERROR);
}

#endif /* __INCsysSystemAce */
